definition module Football

/**	The Football data type and a number of access functions.
*/

import Footballer, randomstream

::	FootballState	= Free     !Football
					| GainedBy !FootballerID
::	Football		= { ballPos		:: !Position3D
					  , ballSpeed	:: !Speed3D
					  }
::	BounceDirection = Down | Up | Forward | Back

instance zero Football

/*	mkFootball returns a football with 3D dimensions.
*/
mkFootball			:: !Position !Speed -> Football

/*	ballIsFree yields True iff argument is (Free ...).
*/
ballIsFree			:: !FootballState -> Bool

/*	ballIsGainedBy yields True iff the ball is in possession by the given player.
*/
ballIsGainedBy		:: !FootballerID !FootballState -> Bool

/*	ballAtCenter returns a non-moving football at the center of the football field.
*/
ballAtCenter		:: !FootballField -> Football

/*	getFootball returns the football (containing its position and speed-information)
	that is either free or gained by a footballer.
	For this reason, the list of footballers must contain all footballers, otherwise
	this function may fail.
*/
getFootball			:: !FootballState !.[Footballer] -> Football

/*	Returns True if the ball is held by a Keeper in his own penaltyarea
	Returns False when the ball is held by a Keeper in open field
	Returns False when the ball is not held by a Keeper
	Keepers should be numbered with 1.
*/
ballGainedByKeeper	:: !FootballState ![Footballer] !ClubName !Home !FootballField -> Bool

/*	Function used for giving a new random direction towards the given BounceDirection (#param1)
*/
bounceBall			:: !BounceDirection !(!Speed3D,!RandomSeed) -> (!Speed3D,!RandomSeed)

radius_football		:== 0.3			// officially it should be 0.113m, but that turns out to be too small for rendering
surface_resistance	:== 0.85		// maximum speed of ball when moving over surface
air_resistance		:== 0.95		// maximum speed of ball when moving through air (should depend on velocity)
accelleration_sec	:== 9.81		// acceleration difference per square second
