implementation module Footballer

import StdEnvExt
import Football, FootballerFunctions, Geometry

instance == Edge where 
	(==) North North				= True
	(==) South South				= True
	(==) _     _					= False
instance == FootballerID where 
	(==) i1 i2  					= i1.clubName == i2.clubName && i1.playerNr == i2.playerNr
instance == Half where 
	(==) FirstHalf  FirstHalf		= True
	(==) SecondHalf SecondHalf		= True
	(==) _          _				= False
instance == Reprimand where 
	(==) Warning    Warning			= True
	(==) YellowCard YellowCard		= True
	(==) RedCard    RedCard			= True
	(==) _          _				= False
instance == Skill where 
	(==) s1 s2						= toString s1 == toString s2
instance == Success where
	(==) Success    Success			= True
	(==) Fail       Fail			= True
	(==) _          _				= False
instance == Footballer where 
	(==) fb1 fb2					= fb1.playerID == fb2.playerID
instance == Home where 
	(==) West       West			= True
	(==) East       East			= True
	(==) _          _				= False
instance == FeintDirection where 
	(==) FeintLeft  FeintLeft		= True
	(==) FeintRight FeintRight		= True
	(==) _          _				= False
instance == FootballerAction where
	(==) (Move     speed1 angle1) (Move     speed2 angle2)	= speed1 == speed2 && angle1 == angle2
	(==) GainBall                 GainBall					= True
	(==) (KickBall speed3D1)      (KickBall speed3D2)		= speed3D1 == speed3D2
	(==) (HeadBall speed3D1)      (HeadBall speed3D2)		= speed3D1 == speed3D2
	(==) (Feint    fd1)           (Feint    fd2)			= fd1 == fd2
	(==) (Tackle   tf1 v1)        (Tackle   tf2 v2)			= tf1 == tf2 && v1 == v2
	(==) Schwalbe                 Schwalbe					= True
	(==) CatchBall                CatchBall					= True
	(==) PlayTheater              PlayTheater				= True
	(==) _ _												= False
instance toString Edge where 
	toString North					= "North"
	toString South					= "South"
instance toString FootballerID where 
	toString i  					= "{clubName=" <+++ i.clubName +++ ",playerNr=" <+++ i.playerNr +++ "}"
instance toString Half where 
	toString FirstHalf				= "FirstHalf"
	toString SecondHalf				= "SecondHalf"
instance toString Reprimand where 
	toString Warning				= "Warning"
	toString YellowCard				= "YellowCard"
	toString RedCard				= "RedCard"
instance toString Skill where 
	toString Running				= "Running"
	toString Dribbling				= "Dribbling"
	toString Rotating				= "Rotating"
	toString Gaining				= "Gaining"
	toString Kicking				= "Kicking"
	toString Heading				= "Heading"
	toString Feinting				= "Feinting"
	toString Jumping				= "Jumping"
	toString Catching				= "Catching"
	toString Tackling				= "Tackling"
	toString Schwalbing				= "Schwalbing"
	toString PlayingTheater			= "PlayingTheater"
instance toString Success where
	toString Success				= "Success"
	toString Fail					= "Fail"
instance toString Home where 
	toString West					= "West"
	toString East					= "East"
instance toString FeintDirection where 
	toString FeintLeft				= "FeintLeft"
	toString FeintRight				= "FeintRight"
instance toString FootballerAction where 
	toString (Move     speed angle)	= "(Move "     <+++ speed  <+++ " " <+++ angle <+++ ")"
	toString  GainBall				= "GainBall"
	toString (KickBall speed)		= "(KickBall " <+++ speed  <+++ ")"
	toString (HeadBall speed)		= "(HeadBall " <+++ speed  <+++ ")"
	toString (Feint    fd)			= "(Feint "    <+++ fd     <+++ ")"
	toString (Tackle   fbID v)		= "(Tackle "   <+++ fbID.playerNr <+++ " " <+++ v <+++ ")"
	toString  Schwalbe				= "Schwalbe"
	toString  CatchBall				= "CatchBall"
	toString  PlayTheater			= "PlayTheater"

defaultFootballer :: !FootballerID -> Footballer
defaultFootballer playerID
	= { playerID		= playerID
	  , name			= "default"
	  , length			= 1.6
	  , pos				= zero 
	  , speed			= zero
	  , nose			= zero
	  , skills			= (Running, Kicking, Dribbling)
	  , effect			= Nothing
	  , stamina			= max_stamina
	  , health			= max_health
	  , brain			= {m=Void, ai=returnAI (Move zero zero)}
	  }

inRadiusOfFootballer						:: !Position !Footballer -> Bool
inRadiusOfFootballer pos player				= inRadiusOfPosition {zero & pxy=pos} xWidthFootballer yWidthFootballer player.length player.pos

skillsAsList								:: !Footballer -> [Skill]
skillsAsList fb								= (\(a,b,c)->[a,b,c]) fb.skills

instance nameOf   FootballerID where nameOf {clubName}	= clubName
instance sameClub FootballerID where sameClub id1 id2	= nameOf id1 == nameOf id2
instance sameClub Footballer   where sameClub fb1 fb2	= sameClub fb1.playerID fb2.playerID

identify_player								:: !FootballerID !Footballer -> Bool
identify_player id fb						= id == fb.playerID

player_identity								:: !Footballer -> FootballerID
player_identity fb							= fb.playerID

getClubName									:: !Footballer -> ClubName
getClubName fb								= nameOf fb.playerID

isKeeper									:: !Footballer -> Bool
isKeeper fb									= fb.playerID.playerNr == 1

isFielder									:: !Footballer -> Bool
isFielder fb								= not (isKeeper fb)

instance nameOf      Footballer where nameOf {name,nose}= name
instance getPosition Footballer where getPosition fb	= fb.pos

/**	Footballer attribute dependent abilities:
*/
maxGainReach								:: !Footballer -> Metre
maxGainReach fb								= (if (isMember Gaining (skillsAsList fb)) 0.5 0.3) * fb.length

maxJumpReach								:: !Footballer -> Metre
maxJumpReach fb								= (if (isMember Jumping (skillsAsList fb)) 0.6 0.4) * fb.length

maxGainVelocityDifference					:: !Footballer !Metre -> Velocity
maxGainVelocityDifference fb d_player_ball	= (if (isMember Gaining (skillsAsList fb)) 15.0 10.0) - distanceDifficulty
where
	distanceDifficulty						= max zero (((0.8*fb.length)^4.0)*(d_player_ball/fb.length))

maxCatchVelocityDifference					:: !Footballer !Metre -> Velocity
maxCatchVelocityDifference fb d_player_ball	= (if (isMember Gaining (skillsAsList fb)) 20.0 17.0) - distanceDifficulty
where
	distanceDifficulty						= max zero (((0.8*fb.length)^4.0) * (d_player_ball/fb.length))

maxKickReach								:: !Footballer -> Metre
maxKickReach fb								= (if (isMember Kicking (skillsAsList fb)) 0.6 0.4) * fb.length

maxHeadReach								:: !Footballer -> Metre
maxHeadReach fb								= (if (isMember Heading (skillsAsList fb)) 0.4 0.2) * fb.length

maxCatchReach								:: !Footballer -> Metre		// includes horizontal jumping
maxCatchReach fb							= (if (isMember Catching (skillsAsList fb)) 1.8 1.5) * fb.length

maxTackleReach								:: !Footballer -> Metre
maxTackleReach fb							= (if (isMember Tackling (skillsAsList fb)) 0.33 0.25) * fb.length

maxVelocityBallKick							:: !Footballer -> Velocity	
maxVelocityBallKick fb						= (if (isMember Kicking (skillsAsList fb)) 27.0 25.0 + fb.length/2.0) * (0.2*fatHealth+0.8)
where
	fatHealth								= getHealthStaminaFactor fb.health fb.stamina

maxVelocityBallHead							:: !Footballer !Velocity -> Velocity
maxVelocityBallHead fb ballSpeed			= 0.7*ballSpeed + (if (isMember Heading (skillsAsList fb)) 7.0 5.0)*(0.1*fatHealth+0.9)
where
	fatHealth								= getHealthStaminaFactor fb.health fb.stamina
	
maxKickingDeviation							:: !Footballer -> Angle
maxKickingDeviation skills					= pi/2.0//if (isMember Kicking skills) (pi/18.0) (pi/2.0)

maxHeadingDeviation							:: !Footballer -> Angle
maxHeadingDeviation skills					= pi/4.0//if (isMember Heading skills) (pi/16.0) (pi/5.0)

maxRotateAngle								:: !Footballer -> Angle
maxRotateAngle fb=:{speed,length}			= pi/18.0*((5.0/speed.velocity)*(length/2.0))

maxFeintStep								:: !Footballer -> Metre
maxFeintStep fb								= if (isMember Feinting (skillsAsList fb)) 0.75 0.5

::	HealthStaminaFactor						:== Real	 	// combination of stamina and health

getHealthStaminaFactor						:: !Health !Stamina -> HealthStaminaFactor
getHealthStaminaFactor health stamina
	| stamina <= health						= stamina
	| otherwise								= avg [stamina,health]

isMove :: !FootballerAction -> Bool
isMove (Move _ _)							= True
isMove _									= False

isGainBall :: !FootballerAction -> Bool
isGainBall GainBall							= True
isGainBall _								= False

isKickBall :: !FootballerAction -> Bool
isKickBall (KickBall _)						= True
isKickBall _								= False

isHeadBall :: !FootballerAction -> Bool
isHeadBall (HeadBall _)						= True
isHeadBall _								= False

isFeint :: !FootballerAction -> Bool
isFeint (Feint _)							= True
isFeint _									= False

isFootballerTackle :: !FootballerAction -> Bool
isFootballerTackle (Tackle _ _)				= True
isFootballerTackle _						= False

isSchwalbe :: !FootballerAction -> Bool
isSchwalbe Schwalbe							= True
isSchwalbe _								= False

isCatchBall :: !FootballerAction -> Bool
isCatchBall CatchBall						= True
isCatchBall _								= False

isPlayTheater :: !FootballerAction -> Bool
isPlayTheater PlayTheater					= True
isPlayTheater _								= False

isSkillOfAction :: !Skill !FootballerAction -> Bool
isSkillOfAction Running        (Move _ _)	= True
isSkillOfAction Rotating       (Move _ _)	= True
isSkillOfAction Gaining        GainBall		= True
isSkillOfAction Kicking        (KickBall _)	= True
isSkillOfAction Heading        (HeadBall _)	= True
isSkillOfAction Feinting       (Feint _) 	= True
isSkillOfAction Tackling       (Tackle _ _)	= True
isSkillOfAction Schwalbing     Schwalbe		= True
isSkillOfAction Catching       CatchBall	= True
isSkillOfAction PlayingTheater PlayTheater	= True
isSkillOfAction _              _			= False

getDefaultField :: FootballField
getDefaultField								= { fwidth = 75.0, flength = 110.0 }

inPenaltyArea :: !FootballField !Home !Position -> Bool
inPenaltyArea field home pos				= north_edge <= pos.py && pos.py <= south_edge && if (home == West) (pos.px <= west_edge) (pos.px >= east_edge)
where
	north_edge								= field.fwidth / 2.0 - radius_penalty_area
	south_edge								= field.fwidth / 2.0 + radius_penalty_area
	west_edge								= penalty_area_depth
	east_edge								= field.flength - penalty_area_depth
	
isWest :: !Home -> Bool
isWest West									= True
isWest _									= False

isEast :: !Home -> Bool
isEast East									= True
isEast _									= False

instance other Home where
	other West								= East
	other East								= West

teamHome :: !ATeam !Half -> Home
teamHome team half
	| team == Team1 && half == FirstHalf || team == Team2 && half == SecondHalf
											= West
	| otherwise								= East

opponentHome :: !ATeam !Half -> Home
opponentHome team half
	| team == Team2 && half == FirstHalf || team == Team1 && half == SecondHalf
											= West
	| otherwise								= East

goal_poles :: !FootballField -> (!Metre,!Metre)
goal_poles field							= (northPole,southPole)
where
	northPole								= (field.fwidth-goal_width)/2.0
	southPole								= northPole + goal_width

isActionOnBall :: !FootballerAction -> Bool
isActionOnBall GainBall						= True
isActionOnBall CatchBall					= True
isActionOnBall (KickBall _)					= True
isActionOnBall (HeadBall _)					= True
isActionOnBall _							= False

isMoved :: !FootballerEffect -> Bool
isMoved (Moved _ _)							= True
isMoved _									= False

isGainedBall :: !FootballerEffect -> Bool
isGainedBall (GainedBall _)					= True
isGainedBall _								= False

isKickedBall :: !FootballerEffect -> Bool
isKickedBall (KickedBall _)					= True
isKickedBall _								= False

isHeadedBall :: !FootballerEffect -> Bool
isHeadedBall (HeadedBall _)					= True
isHeadedBall _								= False

isFeinted :: !FootballerEffect -> Bool
isFeinted (Feinted _)						= True
isFeinted _									= False

/*isReprimanded :: !FootballerEffect -> Bool
isReprimanded (Reprimanded _)				= True
isReprimanded _								= False

isScoredGoal :: !FootballerEffect -> Bool
isScoredGoal (ScoredGoal _)					= True
isScoredGoal _								= False*/

isTackled :: !FootballerEffect -> Bool
isTackled (Tackled _ _ _)					= True
isTackled _									= False

isSchwalbed :: !FootballerEffect -> Bool
isSchwalbed Schwalbed						= True
isSchwalbed _								= False

isCaughtBall :: !FootballerEffect -> Bool
isCaughtBall (CaughtBall _)					= True
isCaughtBall _								= False

isPlayedTheater :: !FootballerEffect -> Bool
isPlayedTheater PlayedTheater				= True
isPlayedTheater _							= False

isOnTheGround :: !FootballerEffect -> Bool
isOnTheGround (OnTheGround _)				= True
isOnTheGround _								= False

failFootballerAction :: !FootballerAction -> FootballerEffect
failFootballerAction (Move   s a)			= Moved    s a
failFootballerAction GainBall				= GainedBall Fail
failFootballerAction CatchBall				= CaughtBall Fail
failFootballerAction (KickBall v)			= KickedBall Nothing
failFootballerAction (HeadBall v)			= HeadedBall Nothing
failFootballerAction (Feint    d)			= Feinted d
failFootballerAction (Tackle p v)			= Tackled p v Fail
failFootballerAction Schwalbe				= Schwalbed
failFootballerAction PlayTheater			= PlayedTheater
failFootballerAction _						= abort "failFootballerAction: unknown action failed"


isReprimandPlayer :: !RefereeAction -> Bool
isReprimandPlayer (ReprimandPlayer _ _)		= True
isReprimandPlayer _							= False

isHands :: !RefereeAction -> Bool
isHands (Hands _)							= True
isHands _									= False

isTackleDetected :: !RefereeAction -> Bool
isTackleDetected (TackleDetected _)			= True
isTackleDetected _							= False

isSchwalbeDetected :: !RefereeAction -> Bool
isSchwalbeDetected (SchwalbeDetected _)		= True
isSchwalbeDetected _						= False

isTheaterDetected :: !RefereeAction -> Bool
isTheaterDetected (TheaterDetected _)		= True
isTheaterDetected _							= False

isDangerousPlay :: !RefereeAction -> Bool
isDangerousPlay (DangerousPlay _)			= True
isDangerousPlay _							= False

isGameOver :: !RefereeAction -> Bool
isGameOver GameOver							= True
isGameOver _								= False

isPauseGame :: !RefereeAction -> Bool
isPauseGame PauseGame						= True
isPauseGame _								= False

isAddTime :: !RefereeAction -> Bool
isAddTime (AddTime _)						= True
isAddTime _									= False

isEndHalf :: !RefereeAction -> Bool
isEndHalf EndHalf							= True
isEndHalf _									= False

isGoal :: !RefereeAction -> Bool
isGoal (Goal _)								= True
isGoal _									= False

isOffside :: !RefereeAction -> Bool
isOffside (Offside _)						= True
isOffside _									= False

isDirectFreeKick :: !RefereeAction -> Bool
isDirectFreeKick (DirectFreeKick _ _ )		= True
isDirectFreeKick _							= False

isGoalKick :: !RefereeAction -> Bool
isGoalKick (GoalKick _)						= True
isGoalKick _								= False

isCorner :: !RefereeAction -> Bool
isCorner (Corner _ _)						= True
isCorner _									= False

isThrowIn :: !RefereeAction -> Bool
isThrowIn (ThrowIn _ _)						= True
isThrowIn _									= False

isPenalty :: !RefereeAction -> Bool
isPenalty (Penalty _)						= True
isPenalty _									= False

isCenterKick :: !RefereeAction -> Bool
isCenterKick (CenterKick _)					= True
isCenterKick _								= False

isAdvantage :: !RefereeAction -> Bool
isAdvantage (Advantage _)					= True
isAdvantage _								= False

isOwnBallIllegally :: !RefereeAction -> Bool
isOwnBallIllegally (OwnBallIllegally _)		= True
isOwnBallIllegally _						= False

isDisplacePlayers :: !RefereeAction -> Bool
isDisplacePlayers (DisplacePlayers _)		= True
isDisplacePlayers _							= False

isContinueGame :: !RefereeAction -> Bool
isContinueGame ContinueGame					= True
isContinueGame _							= False

isTellMessage :: !RefereeAction -> Bool
isTellMessage (TellMessage _)				= True
isTellMessage _								= False

isGoal4ATeam :: !ATeam !RefereeAction -> Bool
isGoal4ATeam t (Goal t`)					= t == t`
isGoal4ATeam _ _							= False


instance other Edge where other North		= South
                          other South		= North
instance other Half where other FirstHalf	= SecondHalf
                          other SecondHalf	= FirstHalf

isFirstHalf :: !Half -> Bool
isFirstHalf FirstHalf						= True
isFirstHalf _								= False

isSecondHalf :: !Half -> Bool
isSecondHalf SecondHalf						= True
isSecondHalf _								= False

getKickPos :: !FootballField !Half !RefereeAction -> Maybe Position
getKickPos field half (GoalKick team)		= Just { py = field.fwidth/2.0
											       , px = if (team == Team1 && half == FirstHalf || team == Team2 && half == SecondHalf) 
										    	             penalty_area_depth
											                 field.flength - penalty_area_depth
											       }
getKickPos field half (Corner team edge)	= Just { px = if (team == Team1 && half == SecondHalf || team == Team2 && half == FirstHalf)
											                 half_radius_corner_kick_area
											                 (field.flength - half_radius_corner_kick_area)
											       , py = if (edge == North)
											                 half_radius_corner_kick_area
											                 (field.fwidth - half_radius_corner_kick_area)
											       }		
where
	half_radius_corner_kick_area			= radius_corner_kick_area / 2.0
getKickPos field half (Penalty team)		= Just { py = field.fwidth/2.0
											       , px = if (team == Team1 && half == SecondHalf || team == Team2 && half == FirstHalf)
											                 penalty_spot_depth
											                 (field.flength - penalty_spot_depth)
											       }
getKickPos field _ (CenterKick _)			= Just { px = field.flength/2.0
											       , py = field.fwidth /2.0
											       }
getKickPos _ _ (DirectFreeKick _ pos)		= Just pos
getKickPos _ _ (ThrowIn _ pos)				= Just pos
getKickPos _ _ _							= Nothing
