implementation module Referee

import matchGame
from   Ivanov import ivanovReferee

allAvailableReferees	:: [FootballField -> Referee]
allAvailableReferees	= [ ivanovReferee ]

instance nameOf Referee where nameOf {Referee | name} = name

defaultReferee		:: Referee
defaultReferee			= { name			= "Default"
						  , brain			= {m = Void,ai = \_ _ _ _ _ _ st -> ([ContinueGame],st)}
						  , refActionPics	= []
						  }

defaultImage			:: !Match !RefereeAction !*env -> (!Bitmap,!*env)	| FileSystem env
defaultImage match rev env
	# bitmapf			= case rev of
							(ReprimandPlayer _ Warning)		= "ivanov_warning.bmp"
							(ReprimandPlayer _ YellowCard)	= "ivanov_yellow.bmp"
							(ReprimandPlayer _ RedCard)		= "ivanov_red.bmp"
							(Hands            _)			= "hands.bmp"
							(OwnBallIllegally _)			= "ivanov_badluck.bmp"
							(TellMessage      _)			= "ivanov_look.bmp"
							(DirectFreeKick   _ p)			= if (p.px < match.theField.flength/2.0) "ivanov_wijst_links.bmp" "ivanov_wijst_rechts.bmp"
							(GoalKick           t)			= if (point_left match t)                "ivanov_wijst_links.bmp" "ivanov_wijst_rechts.bmp"
							(Corner           t _)			= if (point_left match t)                "ivanov_wijst_links.bmp" "ivanov_wijst_rechts.bmp"
							(ThrowIn          t _)			= if (point_left match t)                "ivanov_wijst_links.bmp" "ivanov_wijst_rechts.bmp"
							(Penalty            t)			= if (point_left match t)                "ivanov_wijst_links.bmp" "ivanov_wijst_rechts.bmp"
							(Advantage        _)			= "ivanov_badluck.bmp"
							(TheaterDetected  _)			= "ivanov_theater.bmp"
							_								= "ivanov_fluit.bmp"
	= case openBitmap ("afbeeldingen\\"+++bitmapf) env of
		(Just bm,env)	= (bm,env)
		nothing			= abort "defaultImage: unable to load default picture.\n"
where
	point_left match t	= match.playingHalf == FirstHalf && t == Team1 || match.playingHalf == SecondHalf && t == Team2

defaultSoundFile		:: !RefereeAction -> Maybe String
defaultSoundFile rev	= if (soundfilename == "") Nothing (Just ("sound\\"+++soundfilename))
where
	soundfilename		= defaultSoundFileName rev
	
	defaultSoundFileName (Hands _)					= "stopBecauseOfFoul.wav"
	defaultSoundFileName (TheaterDetected _)		= "tackles_ed.wav"
	defaultSoundFileName (TackleDetected _)			= "tackles_ed.wav"
	defaultSoundFileName (SchwalbeDetected _)		= "tackles_ed.wav"
	defaultSoundFileName (DangerousPlay _)			= "tackles_ed.wav"
	defaultSoundFileName GameOver					= "endGameOrHalf.wav"
	defaultSoundFileName EndHalf					= "endGameOrHalf.wav"
	defaultSoundFileName (Offside _)				= "offside.wav"
	defaultSoundFileName (GoalKick _)				= "ballOut.wav"
	defaultSoundFileName (Corner _ _)				= "ballOut.wav"
	defaultSoundFileName (ThrowIn _ _)				= "ballOut.wav"
	defaultSoundFileName (Goal _)					= "CenterKick.wav"
	defaultSoundFileName (OwnBallIllegally _)		= "wrongPosition2restartFrom.wav"
	defaultSoundFileName _							= ""

randomlessRefereeAI		:: (RefereeAI memory) -> RefereeAI (memory,RandomSeed)
randomlessRefereeAI brainf
	= randomless brainf
where
	randomless brainf playingtime unit maybe_ball half team1 team2 (memory,seed)
		# (decisions,memory)	= brainf playingtime unit maybe_ball half team1 team2 memory
		= (decisions,(memory,seed))

amnesiaRefereeAI		:: (RefereeAI RandomSeed) -> RefereeAI (memory,RandomSeed)
amnesiaRefereeAI brainf
	= amnesia brainf
where
	amnesia brainf playingtime unit maybe_ball half team1 team2 (memory,seed)
		# (decisions,seed)		= brainf playingtime unit maybe_ball half team1 team2 seed
		= (decisions,(memory,seed))

witlessRefereeAI		:: (RefereeAI Void) -> RefereeAI (memory,RandomSeed)
witlessRefereeAI brainf
	= witless brainf
where
	witless brainf playingtime unit maybe_ball half team1 team2 state
		= (fst (brainf playingtime unit maybe_ball half team1 team2 Void),state)

