implementation module matchGame

import StdEnvExt
import guiInterface, matchControl

makeResult :: !Tournament -> TournamentResult
makeResult t	= sortBy tournamentSort (makeResult` t [])
where
	makeResult` :: !Tournament !TournamentResult -> TournamentResult
	makeResult` [] result 		= result
	makeResult` [(t1,t2,(s1,s2)):tx] result
		# (teams,field)			= (allAvailableTeams,getDefaultField)
		# (westNames,eastNames) = (map nameOf [teamF West field \\ teamF <- teams], map nameOf [teamF East field \\ teamF <- teams])		
		# otherName1			= getFromOtherList (nameOf t1) (westNames,eastNames)
		# otherName2			= getFromOtherList (nameOf t2) (westNames,eastNames)
		| s1 > s2 				= makeResult` tx (addPoints 3 (nameOf t1,otherName1) (addPoints 0 (otherName2,nameOf t2) result))
		| s1 == s2				= makeResult` tx (addPoints 1 (nameOf t1,otherName1) (addPoints 1 (otherName2,nameOf t2) result))
		| otherwise				= makeResult` tx (addPoints 3 (otherName2,nameOf t2) (addPoints 0 (nameOf t1,otherName1) result))
	where
		addPoints :: !Points !(!ClubName,!ClubName) !TournamentResult -> TournamentResult
		addPoints p (name1,name2) []
			= [(name1,p)]
		addPoints p (name1,name2) [t1=:(tr_team,tr_points):trls]			 
			| name1 == tr_team || name2 == tr_team
				= [(tr_team,tr_points + p):trls]
			| otherwise
				= [t1:addPoints p (name1,name2) trls]
	
	tournamentSort :: !(a,!Points) !(a,!Points) -> Bool
	tournamentSort (_,p1) (_,p2)	= p1 > p2
	
instance toString Tournament where
	toString []				= ""
	toString tournament		= printLine 49 <+++ printTournament tournament
	where
		printTournament :: !Tournament -> String
		printTournament []	= printLine 49
		printTournament [(t1,t2,(ws,es)):ts] 
			= start <+++ space <+++ end <+++ printTournament ts
		where
			start	 		= "| " <+++ nameOf t1 <+++ " VS " <+++ nameOf t2 <+++ ":"
			space 			= addSpaces (40 - size start)
			end				= westScore <+++ spaceScore1 <+++ "-" <+++ spaceScore2 <+++ eastScore <+++ " |\n"
			westScore		= toString ws
			eastScore		= toString es
			spaceScore1		= addSpaces (3 - (size westScore))
			spaceScore2		= addSpaces (3 - (size eastScore))
			
instance toString TournamentResult where
	toString []				= printLine 32
	toString [(team,points):trls]
		= start <+++ space <+++ pointStr <+++ space2 <+++ end <+++ trls
	where
		start				= "| " <+++ team
		space				= addSpaces (20 - (size start))
		pointStr			= toString points
		space2				= addSpaces (3 - (size pointStr))
		end					= " punten |\n"
							
addSpaces :: !Int -> String
addSpaces n					= showList "" (repeatn n " ")

printLine :: !Int -> String
printLine 0					= "\n"
printLine i					= " " <+++ printLine` (twice dec i) 
where
	printLine` :: !Int -> String
	printLine` 0			= " \n"
	printLine` i			= "-" <+++ printLine` (dec i)

incFrames :: !FootballGame -> FootballGame
incFrames game=:{frames}	= {game & frames=frames+1}
