implementation module matchLog

import StdEnvExt, fileIO
import matchControl

logFile :: String
logFile = "SoccerFun.log"

getWhatToLog :: !*env -> (!WhatToLog,!*env) | FileSystem env
getWhatToLog env
	# (open,file,env)		= fopen fileStr FReadText env
	| not open				= (defaultLogging,env)
	# (line,file)			= freadline file
	# logging				= fromString line
	# (closed,env)			= fclose file env
	| not closed			= abort ("matchControl,getLogging: unable to close file "+++fileStr+++"\n")
	| otherwise				= (logging,env)		
where
	fileStr					= "logMatch.conf"

instance toString WhatToLog where 
	toString l = "{footballerActions=" <+++ l.footballerActions +++ 
				 ",fbPositions="       <+++ l.fbPositions       +++ 
				 ",refEvents="         <+++ l.refEvents         +++ 
				 ",ballPosition="      <+++ l.ballPosition      +++ 
				 "}"
instance fromString WhatToLog where
	fromString str
		# (okActions,  actions,  str)			= isBool (literal "{footballerActions=" str)
		# (okPositions,positions,str)			= isBool (literal ",fbPositions="       str)
		# (okRefEvents,events,   str)			= isBool (literal ",refEvents="         str)
		# (okPositions,pos,      str)			= isBool (literal ",ballPosition="      str)
		| okActions && okPositions && okRefEvents && okPositions && str=="}"
			= { footballerActions				= actions
			  , fbPositions						= positions
			  , refEvents						= events
			  , ballPosition					= pos
			  }
		| otherwise
			= defaultLogging
	where
		literal literal str
			| str%(0,size literal-1) == literal	= str%(size literal,size str-1)
			| otherwise							= str
		isBool str
			| str%(0,3) == "True"				= (True,True, str%(4,size str-1))
			| str%(0,4) == "False"				= (True,False,str%(5,size str-1))
			| otherwise							= (False,undef,str)

defaultLogging		= { footballerActions		= False
					  , fbPositions				= False
					  , refEvents				= False
					  , ballPosition			= False
					  }


/*	logMatch logging match succeededActions refereeActions env
		writes the log fields indicated in logging from the match and actions to disk.
*/
logMatch :: !WhatToLog !Match ![RefereeAction] ![FootballerWithAction] !*env -> *env | FileSystem env
logMatch logging match_to_log=:{team1,team2,playingTime,theBall} refereeActions succeededActions env
	| logging.footballerActions || logging.fbPositions || logging.refEvents || logging.ballPosition
		# env				= writeFile logFile ("BEGIN STEP "<+++playingTime<+++"\n") env
		# env				= if (logging.footballerActions) (writeFile logFile (logActions    allPlayers succeededActions) env) env
		# env				= if (logging.fbPositions)       (writeFile logFile (logPositions  allPlayers)                  env) env
		# env				= if (logging.refEvents)         (writeFile logFile (logRefereeActions refereeActions)          env) env
		# env				= if (logging.ballPosition)      (writeFile logFile ("BallPos: " <+++ ball.ballPos+++"\n")      env) env
		# env				= writeFile logFile ("END STEP "<+++playingTime<+++"\n") env
		= env
	| otherwise
		= env
where
	allPlayers				= team1 ++ team2
	ball					= getFootball theBall allPlayers 

	logActions :: ![Footballer] ![FootballerWithAction] -> String
	logActions _ []		= ""
	logActions allPlayers [(action,playerID=:{clubName,playerNr}):mfals]
						= "FootballerWithAction\n\tTeam: "     <+++ 
						  clubName                             <+++ 
						  "\n\tNumber: "                       <+++ 
						  playerNr                             <+++ 
						  "\n\tName: "                         <+++ 
						  nameOf self                          <+++ 
						  "\n\tAction: "                       <+++ 
						  improveString (toString action)      <+++ 
						  "\n"                                 <+++ logActions allPlayers mfals
	where
		self			= find1 (identify_player playerID) allPlayers
		
	improveString		= (replaceInString "}\"" "}") o (replaceInString "\"{" "{") o (replaceInString "\\\"" "\"")

	logPositions :: ![Footballer] -> String
	logPositions fbs	= "POSITIONS:\n" <+++ foldl printFootballer "" fbs
	where
		printFootballer str {playerID={clubName,playerNr},name,pos}
						= str <+++ "\t(" <+++ clubName <+++ ",(" <+++ playerNr <+++"," <+++ name <+++")):\t" <+++ pos <+++ "\n"

	logRefereeActions :: ![RefereeAction] -> String
	logRefereeActions []				= ""
	logRefereeActions refls				= "REFEREE_ACTIONS:" <+++ logRefereeActions` refls <+++ "\n"
	where
		logRefereeActions` :: ![RefereeAction] -> String
		logRefereeActions` []			= ""
		logRefereeActions` [ref:refls]	= "\n\t" <+++ logRefereeAction ref <+++ logRefereeActions` refls
		
		logRefereeAction :: !RefereeAction -> String
		logRefereeAction (ReprimandPlayer   tfp r)	= "(ReprimandPlayer " <+++ tfp.playerNr <+++ " " <+++ r <+++ ")"
		logRefereeAction (Hands 			tfp)	= "(Hands "           <+++ tfp.playerNr <+++ ")"
		logRefereeAction (TackleDetected	tfp)	= "(TackleDetected "  <+++ tfp.playerNr <+++ ")"
		logRefereeAction (SchwalbeDetected	tfp)	= "(SchwalbeDetected "<+++ tfp.playerNr <+++ ")"
		logRefereeAction (TheaterDetected 	tfp)	= "(TheaterDetected " <+++ tfp.playerNr <+++ ")"
		logRefereeAction (DangerousPlay		tfp)	= "(DangerousPlay "   <+++ tfp.playerNr <+++ ")"
		logRefereeAction  GameOver					= "GameOver"
		logRefereeAction  PauseGame					= "PauseGame"
		logRefereeAction (AddTime 			t)		= "(AddTime "         <+++ showMinutes t    <+++ ")"
		logRefereeAction  EndHalf					= "EndHalf"
		logRefereeAction (Goal 				t)		= "(Goal "            <+++ t <+++ ")"
		logRefereeAction (Offside 			tfp)	= "(Offside "         <+++ tfp.playerNr <+++ ")"
		logRefereeAction (DirectFreeKick   	t p)	= "(DirectFreeKick "  <+++ t <+++ " " <+++ p <+++ ")"
		logRefereeAction (GoalKick         	t)		= "(GoalKick "        <+++ t <+++ ")"
		logRefereeAction (Corner	      	t e)	= "(Corner "          <+++ t <+++ " " <+++ e <+++ ")"
		logRefereeAction (ThrowIn          	t p)	= "(ThrowIn "         <+++ t <+++ " " <+++ p <+++ ")"
		logRefereeAction (Penalty		 	t)		= "(Penalty "         <+++ t <+++ ")"
		logRefereeAction (CenterKick		t)		= "(CenterKick "      <+++ t <+++ ")"
		logRefereeAction (Advantage			t)		= "(Advantage "       <+++ t <+++ ")"
		logRefereeAction (OwnBallIllegally  tfp)	= "(OwnBallIllegally" <+++ tfp.playerNr <+++ ")"
		logRefereeAction (DisplacePlayers	ds)		= "DisplacePlayers"
		logRefereeAction  ContinueGame				= "ContinueGame"
		logRefereeAction (TellMessage		txt)	= "(TellMessage "     <+++ txt <+++ ")"
