implementation module Gui2D

import StdEnvExt, StdIOExt
import fileIO
import digitdisplay, textdisplay, renderGame

SoccerFunGUI2D :: !*World -> *World
SoccerFunGUI2D world
	# (matchAtStart,world) 	= getBeginMatch world
	# (theGame, world) 		= acquireGame matchAtStart {refereePics = []} world
	# (ids,     world)		= openIds 3 world
	# (displIds,world)		= sseqList (repeatn 2 openDigitDisplayId) world
	# (teamdIds,world)		= sseqList (repeatn 5 openTextDisplayId)  world
	# (frameId, world)		= openTextDisplayId world
	= startIO SDI theGame
	              (initGUI (ids,displIds,teamdIds,frameId))
	              [ProcessClose closeProcess]
	              world

getBeginMatch :: !*env -> (!Match,!*env) | TimeEnv, FileSystem env
getBeginMatch env
	| isEmpty teams		= abort "There are no teams.\nPlease add teams to \"allAvailableTeams\" in the \"Team\"-module\n"	
	| isEmpty referees	= abort "There are no referees.\nPlease add referees to \"allAvailableReferees\" in the \"Referee\"-module\n"
	| otherwise			= getMatchStart (hd teams West field) (hd teams East field) getDefaultField (hd referees field) defaultPlayingTime env
where
	teams				= allAvailableTeams
	referees			= allAvailableReferees
	field				= getDefaultField

frameRatePrefix			:== "Rounds/sec: "

showTime :: !Minutes -> String
showTime minutes		= toString (seconds/60) <+++ ":" <+++ if (seconds mod 60 < 10) "0" "" <+++ (seconds mod 60) <+++" min"
where
	seconds				= toInt (((toReal (toInt (minutes * 100.0)))/100.0) * 60.0)

showHalf :: !Half -> String
showHalf FirstHalf		= "1st half"
showHalf SecondHalf		= "2nd half"

initGUI :: !(![Id],![DigitDisplayId],![TextDisplayId],!TextDisplayId) !(PSt FootballGame) -> (PSt FootballGame)	
initGUI ([timerId,fieldId,windowId],[westId,eastId],[team1Id,team2Id,halfId,timeId,refId],framesId) pSt=:{ls=ls=:{match=matchAtStart,options}}
	# (penAtts,pSt)		= accPIO renderAttributes pSt
	# (penFont,pSt)		= accPIO (accScreenPicture (openFont {fName="Courier New",fSize=32,fStyles=[BoldStyle]})) pSt
	# (frameFont,pSt)	= accPIO (accScreenPicture (openFont {fName="Courier New",fSize=16,fStyles=[BoldStyle]})) pSt
	# (underFont,pSt)	= accPIO (accScreenPicture (openFont {fName="Courier New",fSize=12,fStyles=[BoldStyle]})) pSt
	# refsName			= nameOf matchAtStart.theReferee
	# wDef				= Window "SoccerFun: Van Gaal's Electronic Notebook" 
							(	LayoutControl
								(	teamName team1Id (nameOf matchAtStart.team1) penFont WestColour
									:+: DigitDisplay  westId (IntegerFormat 2) digitDisplaySize WestColour []
									:+: DigitDisplay  eastId (IntegerFormat 2) digitDisplaySize EastColour []
									:+: teamName team2Id (nameOf matchAtStart.team2) penFont EastColour
								)	[ ControlPos (Center,zero) ]
								:+: frames framesId frameRatePrefix frameFont Green 
								:+:	CustomControl footballFieldDisplaySize (look ls.match)	[ ControlId    fieldId
																							, ControlMouse (\mst -> case mst of MouseUp _ _ -> True; _ -> False) Able (noLS1 (const nextstep))
																							, ControlPen   penAtts
																							, ControlPos   (Center,zero)
																							, ControlResize (\_ _ wViewSize=:{h} -> {wViewSize & h=h-digitDisplaySize.h-60})
																						]
								:+: TextDisplay halfId (showHalf matchAtStart.playingHalf) {w=80,h=15}
										[ ControlPen [PenColour White,PenBack Black,PenFont (snd underFont)]
										//, ControlResize (\_ _ size -> {size & h=size.h-digitDisplaySize.h})
										, ControlPos (LeftBottom,OffsetVector {vx=10,vy=(~7)})
										]
								:+: TextDisplay timeId ((showTime matchAtStart.playingTime)) {w=80,h=15}
										[ ControlPen [PenColour White,PenBack Black,PenFont (snd underFont)]
										, ControlPos (Center,OffsetVector {vx=zero,vy=zero})
										]
								:+: TextDisplay refId ("Referee: "<+++refsName) {w=80+(12*(size refsName)),h=15}
										[ ControlPen [PenColour White,PenBack Black,PenFont (snd underFont)]
										, ControlPos (RightBottom,OffsetVector {vx=zero,vy=(~7)})//(3*(size "test"))
										]
							)
							[ WindowViewSize	{w=footballFieldDisplaySize.w-1, h=footballFieldDisplaySize.h+digitDisplaySize.h}
							, WindowLook False	stdUnfillNewFrameLook
							, WindowPen			[PenBack Black]
							, WindowId			windowId
							]
	# (error,pSt)		= openWindow undef wDef pSt
	# pSt				= setDigitDisplayValue westId (fst ls.match.score) pSt
	# pSt				= setDigitDisplayValue eastId (snd ls.match.score) pSt
	| error<>NoError	= abort "Could not create window.\n"
	# speed				= intervalFactor ls.options.displaySpeed
	# tDef				= Timer (toInt ((toReal ticksPerSecond)*ls.match.unittime*speed)) NilLS [ TimerFunction    (noLS1 (const nextstep))
								                                                                , TimerId          timerId
								                                                                , TimerSelectState Unable
								                                                                ]
	# (error,pSt)		= openTimer undef tDef pSt
	| error<>NoError	= abort "Could not create step timer.\n"
	# tDef				= Timer ticksPerSecond NilLS [TimerFunction (noLS1 (const frameRate))]
	# (error,pSt)		= openTimer undef tDef pSt
	| error<>NoError	= abort "Could not open frame rate timer.\n"
	# mDef				= Menu "&File"
							(	MenuItem "E&xit"        [MenuFunction (noLS closeProcess),       MenuShortKey 'q']
							)	[]
	# (error,pSt)		= openMenu undef mDef pSt
	| error<>NoError	= abort "Could not create File menu.\n"
	# mDef2				= Menu "&Game"
							(   MenuItem "&Run"         [MenuFunction (noLS continuef),          MenuShortKey 'r']
							:+:	MenuItem "&Step"        [MenuFunction (noLS stepf),              MenuShortKey 's']
							:+:	MenuItem "&Halt"        [MenuFunction (noLS haltf),              MenuShortKey 'h']
							:+: MenuSeparator           []
							:+: SubMenu "&Mode"    ( RadioMenu [ ("&Realistic",  Nothing,Nothing,noLS (restartRealisticf))
							                                   , ("&Predictable",Nothing,Nothing,noLS (restartPredictablef))
							                                   ] 0 []
							                       ) []
							:+: SubMenu "Sp&eed"   ( RadioMenu [ ("&Normal", Nothing, Just '2', noLS (changeSpeedf Normal))
															   , ("&Slow",   Nothing, Just '`', noLS (changeSpeedf Slow))
															   , ("&Target", Nothing, Just '1', noLS (changeSpeedf Target))
															   , ("&Faster", Nothing, Just '3', noLS (changeSpeedf Faster))
															   , ("Fa&stest",Nothing, Just '4', noLS (changeSpeedf Fastest))
															   ] 0 []
												   ) []
							:+: SubMenu "Re&feree" ( RadioMenu [ ("&Show",   Nothing, Just '+', noLS (showReff True))
															   , ("&NoShow", Nothing, Just '-', noLS (showReff False))
															   ] 0 []
												   ) []
							)	[]
	# (error,pSt)		= openMenu undef mDef2 pSt
	| error<>NoError	= abort "Could not create Game menu.\n"
	# mDef3				= Menu "&Referee"
							(	RadioMenu (makeRefereeMenu (apply matchAtStart.theField allAvailableReferees)) 0 []
							) []
	# (error,pSt)		= openMenu undef mDef3 pSt
	| error<>NoError	= abort "Could not create Referee menu.\n"
	# mDef4 			= Menu "Team&1"
							( 	RadioMenu (makeTeamMenu West (getAllTeamsOfHome West matchAtStart.theField)) 0 []
							) []
	# (error,pSt)		= openMenu undef mDef4 pSt
	| error<>NoError	= abort "Could not create Team1 menu.\n"
	# mDef5 			= Menu "Team&2"
							( 	RadioMenu (makeTeamMenu East (getAllTeamsOfHome East matchAtStart.theField)) 0 []
							) []
	# (error,pSt)		= openMenu undef mDef5 pSt
	| error<>NoError	= abort "Could not create Team2 menu.\n"
	# pSt				= appPIO (setWindowViewSize windowId {w=640,h=550}) pSt
	| not options.showSplash
						= pSt
	# (spls,pSt)		= getSplashImageForGui pSt
	# splashDef			= Dialog "The Footballer's Mind"
							(	CompoundControl 
								(	ButtonControl "&Close" [ControlFunction (noLS closeActiveWindow)]
								)
								[	ControlLook True (const2 (drawAt zero spls.img))
								,	ControlViewSize  (getBitmapSize spls.img)
								]
							) []
	# ((error,_),pSt)	= openModalDialog undef splashDef pSt
	| error<>NoError	= abort "Could not open splash screen.\n"
	| otherwise			= pSt
where
	makeTeamMenu :: !Home ![Team] -> [MenuRadioItem *(Match,PSt FootballGame)]
	makeTeamMenu home teams			= [(nameOf t, Nothing, Nothing, noLS (changeteamf home t)) \\ t<-teams]
								
	makeRefereeMenu :: ![Referee] -> [MenuRadioItem *(Match,PSt FootballGame)]
	makeRefereeMenu referees		= [(nameOf r, Nothing, Nothing, noLS (changereff r)) \\ r<-referees]

	digitDisplaySize				= { w=24, h=36 }
	footballFieldDisplaySize		= { w=640,h=400}

	teamName id name penFont colour	= TextDisplay id name (teamSize footballFieldDisplaySize)
										[ ControlPen [PenColour colour,PenBack Black,PenFont (snd penFont)]
										, ControlResize (const2 teamSize)
										]
	where teamSize {w}				= {digitDisplaySize & w=(w-digitDisplaySize.w*4)/2}
	
	frames id name penFont colour	= TextDisplay id name {digitDisplaySize & w=300} 
										[ ControlPen [PenColour colour, PenBack Black, PenFont (snd penFont)]
								        , ControlPos (Center,zero)
										]
		
	look match _					= renderMatch match
		
	stepf :: !(PSt FootballGame) -> PSt FootballGame
	stepf pSt						= nextstep (appPIO (disableTimer timerId) pSt)
		
	showReff :: !Bool !(PSt FootballGame) -> PSt FootballGame
	showReff show pSt				= {pSt & ls = {pSt.ls & options = {pSt.ls.options & noReferee = not show}}}
	
	nextstep :: !(PSt FootballGame) -> PSt FootballGame
	nextstep pSt=:{ls=game,io}
		# (refEvents,game,io)				= stepMatchForGui game io
		# pSt								= {pSt & io = setControlLook fieldId True (False,look game.match) io
											       , ls = incFrames game
											  }
		= analyseRefEvents refEvents pSt
	where
		analyseRefEvents :: ![RefereeAction] !(PSt FootballGame) -> PSt FootballGame
		analyseRefEvents refEvents pSt		= foldl analyseRefEvent pSt refEvents
		where
			analyseRefEvent :: !(PSt FootballGame) !RefereeAction -> PSt FootballGame
			analyseRefEvent pSt=:{ls} rev
				| or (apply rev [isContinueGame,isDisplacePlayers,isDirectFreeKick,isCenterKick,isPauseGame,isAddTime])
											= pSt
				# pSt						= refereeDialog rev pSt
				# pSt						= if (isGameOver rev) (appPIO (disableTimer timerId) pSt) pSt
				| isGoal4ATeam Team1 rev	= setDigitDisplayValue westId (fst ls.match.score) pSt
				| isGoal4ATeam Team2 rev	= setDigitDisplayValue eastId (snd ls.match.score) pSt
				| otherwise					= pSt
			
			refereeDialog :: !RefereeAction !(PSt FootballGame) -> PSt FootballGame
			refereeDialog rev pSt=:{ls=ls=:{match=match=:{theReferee=referee=:{Referee | name}}}}
				| pSt.ls.options.noReferee	= pSt
				# pSt						= case defaultSoundFile rev of
												Just sound	= appPIO (makeSound sound) pSt
												nothing		= pSt
				# pSt						= haltf pSt
				# (image,pSt)				= accPIO (defaultImage match rev) pSt
				# (closeId, pSt)			= openId pSt
				# (dialogId,pSt)			= openId pSt
				# tDef						= Timer (3*ticksPerSecond) NilLS [ TimerFunction (noLS1 (const (continuef o (closeWindow dialogId) o (appPIO (closeTimer closeId)))))
											                                 , TimerId       closeId
											                                 ]
				# (error,  pSt)				= openTimer undef tDef pSt
				| error<>NoError			= abort "Could not open referee dialog timer.\n"
				# refereeDef				= Dialog ("Referee " <+++ name)
												(	TextControl (showSuccintRefereeAction rev) []
												:+: CustomControl (getBitmapSize image) (const2 (drawAt zero image)) []
												)	[WindowId dialogId]
				# ((error,_),pSt)			= openModalDialog undef refereeDef pSt
				| error<>NoError			= abort "Could not open referee dialog.\n"
				| otherwise					= pSt

	changeteamf :: !Home !Team !(PSt FootballGame) -> PSt FootballGame
	changeteamf home team pSt=:{ls=game=:{match},io}
		# match						= {match & theBall     = Free (ballAtCenter match.theField)
									         , playingHalf = FirstHalf
									         , playingTime = defaultPlayingTime
									         , score       = (0,0)
									         , team1       = if (home == West) team match.team1
									         , team2       = if (home == East) team match.team2
									         , theReferee  = getRefereeFresh (nameOf match.theReferee) match.theField
									  }
		# pSt						= {pSt   & ls          = {game & match=match}
									         , io          = setControlLook fieldId True (False,look match) io
									  }
		# pSt						= setTextDisplayText (if (home == West) team1Id team2Id) (nameOf team) pSt
		= pSt

	changereff :: !Referee !(PSt FootballGame) -> PSt FootballGame
	changereff ref=:{Referee | name} pSt
		# pSt						= setTextDisplayText refId ("Referee: " <+++ name) pSt //todo: de positionering moet aangepast worden 
		= appPIO (disableTimer timerId) {pSt & ls={pSt.ls & match = {pSt.ls.match & theReferee = ref}}}
	 
	continuef :: !(PSt FootballGame) -> PSt FootballGame
	continuef pSt					= appPIO (enableTimer timerId) pSt
	
	haltf :: !(PSt FootballGame) -> PSt FootballGame
	haltf pSt						= appPIO (disableTimer timerId) pSt
		
	changeSpeedf :: !DisplaySpeed !(PSt FootballGame) -> PSt FootballGame
	changeSpeedf speed pSt=:{ls}
		# timerInterval 			= toInt ((toReal ticksPerSecond)*ls.match.unittime*(intervalFactor speed))
		# pSt						= {pSt & ls = {FootballGame | ls & options = {Options | ls.options & displaySpeed = speed}}}
		= appPIO (enableTimer timerId o setTimerInterval timerId timerInterval o disableTimer timerId) pSt
		
	restartRealisticf :: !(PSt FootballGame) -> PSt FootballGame
	restartRealisticf pSt=:{ls=ld=:{match}}
									= {pSt & ls={FootballGame | ls & match={match & nextRandomP=nextRandomP}}}
		
	restartPredictablef :: !(PSt FootballGame) -> PSt FootballGame
	restartPredictablef pSt=:{ls=ls=:{match}}
									= {pSt & ls={FootballGame | ls & match={match & nextRandomP=next1}}}

	frameRate :: !(PSt FootballGame) -> PSt FootballGame
	frameRate pSt=:{ls=ls=:{frames}}
		# pSt						= setTextDisplayText timeId (showTime  ls.match.playingTime) pSt
		# pSt						= setTextDisplayText halfId (showHalf ls.match.playingHalf) pSt
		= setTextDisplayText framesId (frameRatePrefix <+++ frames) {pSt & ls={ls & frames=0}}

getRefereeFresh :: !String !FootballField -> Referee
getRefereeFresh refsName field		= hd [r field \\ r <- allAvailableReferees | nameOf (r field) == refsName]

getAllTeamsOfHome :: !Home !FootballField -> [Team]
getAllTeamsOfHome home field		= apply field (apply home allAvailableTeams)

showSuccintRefereeAction :: !RefereeAction -> String
showSuccintRefereeAction refAction
	= case refAction of
			(ReprimandPlayer  id r)	= player id                     <+++ " receives " <+++ r
			(Hands            id)	= "Hands by "                   <+++ player id
			(TackleDetected   id)	= "Tackle by "                  <+++ player id
			(SchwalbeDetected id)	= "Schwalbe by "                <+++ player id
			(TheaterDetected  id)	= "Theater by "                 <+++ player id
			(DangerousPlay    id)	= "Dangerous play by "          <+++ player id
			GameOver				= "Game ended"
			PauseGame				= "Game paused"
			(AddTime  t)			= "Extra time added: "          <+++ showMinutes t
			EndHalf					= "First half ended"
			(Goal     t)			= "Goal for "                   <+++ t
			(Offside        id)		= "Offside by "                 <+++ player id
			(DirectFreeKick t p)	= "Direct free kick for "       <+++ t
			(GoalKick t)			= "Goal kick for "              <+++ t
			(Corner   t e)			= "Corner for "                 <+++ t
			(ThrowIn  t p)			= "Throw in for "               <+++ t
			(Penalty    t)			= "Penalty for "                <+++ t
			(CenterKick t)			= "Center kick for "            <+++ t
			(Advantage t)			= "Advantage for "              <+++ t
			(OwnBallIllegally id)	= "Illegal ball possession by " <+++ player id
			(DisplacePlayers _)		= "Players displaced"
			ContinueGame			= "Game continued"
			(TellMessage txt)		= txt
where
	player {clubName,playerNr}		= clubName <+++"[" <+++ playerNr <+++ "]"
