implementation module guiInterface

import fileIO
import matchGame

inDir img :== "afbeeldingen\\"+++img+++".bmp"

getSplashImageForGui :: !*env -> (!GamePicture,!*env) | FileSystem env
getSplashImageForGui env
	# pathToSplash		= inDir "AmsterdamArenA"
	# (splashImg, env)	= getImage pathToSplash env
	= ({img = splashImg, path = pathToSplash},env)

acquireGame :: !Match !ActionPics !*env -> (!FootballGame,!*env) | TimeEnv, FileSystem env
acquireGame match apls env
	# (whatToLog,env)	= getWhatToLog env
	= ( { match			= match
		, actionPics	= apls
		, history		= {time=2.0,past=[]}
		, frames		= 0
		, options		= { closeReferee	= True		// Peter: currently does not make sense to set to False, as replays have been eliminated
		              	  , showSplash		= False
		                  , displaySpeed	= Normal
		                  , noReferee		= False
		                  , tournament		= Nothing
		                  }
		, logging		= whatToLog
		},env)

stepMatchForGui :: !FootballGame !*env -> (![RefereeAction], !FootballGame, !*env) | FileSystem env
stepMatchForGui game env
	| timeLeft game
		# ((refEvents,actions),match)	= stepMatch game.match
		# env							= logMatch game.logging match refEvents actions env
		= (refEvents,{game & match=match},env)
	| otherwise
		= ([GameOver],game,env)

/* Function description
	*	Parameters: !*Env
					FootballGame
	* 	Returns:	(Score,!*Env)			- the score at the end of the game
	* 	Description:Calculates an entire match and returns the endscore. 
					Usefull for quick calculating without the need to control the match itself
*/
computeEntireMatch :: !FootballGame !*env -> (!Score,!*env) | FileSystem env
computeEntireMatch game env
	| timeLeft game
		# ((refEvents,actions),match)	= stepMatch game.match
		# env							= logMatch game.logging match refEvents actions env
		= computeEntireMatch {game & match=match} env
	| otherwise
		= (game.match.score,env)
	
timeLeft :: !FootballGame -> Bool
timeLeft game = game.match.playingTime > 0.0

defaultPlayingTime :: PlayingTime
defaultPlayingTime = 0.66//1.0

getActionPicsForGui :: !Match !*env -> (!ActionPics,!*env) | FileSystem env
getActionPicsForGui match env
	# (refereePics,env)	= buildToImages match.theReferee.refActionPics env
	= ({refereePics=refereePics},env)
where
	buildToImages :: ![Path] !*env -> (![GamePicture],!*env) | FileSystem env
	buildToImages [] env	= ([],env)	
	buildToImages [l:ls] env
		# (gp,env)			= fillWithPics` l env
		# (ls,env)			= buildToImages ls env
		= ([gp:ls],env)
	where
		fillWithPics` :: !Path !*env -> (!GamePicture,!*env) | FileSystem env
		fillWithPics` path env
			# (img,env) = getImage path env
			= ({img=img, path=path},env)

intervalFactor :: !DisplaySpeed -> Real
intervalFactor Slow		= slow
intervalFactor Target	= target
intervalFactor Normal	= normal
intervalFactor Faster	= faster
intervalFactor Fastest	= fastest

slow					:== 5.0		//= 2 fps
target					:== one 	//=10 fps
normal					:== 0.5 	//=20 fps
faster					:== 0.2 	//=50 fps
fastest					:== zero	//=maximum fps
