implementation module renderGame

import matchControl
import StdIOCommon, StdPSt

/**	renderMatch match updSt
		renders a complete match, given the current dimensions of the drawing
		environment (updSt).
*/
renderMatch					:: !Match !UpdateState !*Picture -> *Picture
renderMatch match=:{theField,team1,team2} updSt=:{newFrame} picture
	# picture				= unfill {zero & corner2=ppix {px=flength,py=fwidth}}                                             picture	// football field
	# picture				= setPenColour Black                                                                              picture
	# picture				= fill {corner1=ppix {px=flength,py=0.0},corner2={x=w,y=pix fwidth}}                              picture	// exterior right
	# picture				= fill {corner1=ppix {px=0.0,py=fwidth}, corner2={x=w,y=h}}                                       picture	// exterior bottom
	# picture				= setPenColour White                                                                              picture
	# picture				= draw {zero & corner2=ppix {px=flength,py=fwidth}}                                               picture	// field outline
	# picture				= drawMiddleArea                                                                                  picture	// middle line, circle and spot
	# picture				= drawGoals                                                                                       picture	// goals
	# picture				= drawGoalAreas                                                                                   picture	// goal areas
	# picture				= drawPenaltyAreas                                                                                picture	// penalty areas
	# picture				= drawCornerAreas                                                                                 picture	// corner areas
	# picture				= foldr (drawPlayer home_colour)                                                            picture team1	// team1 players
	# picture				= foldr (drawPlayer away_colour)                                                            picture team2	// team2 players
	# picture				= setPenColour ball_colour                                                                        picture
	# picture				= fillAt (ppix theBall.ballPos.pxy) (circle (pix ((1.0+theBall.ballPos.pz/5.0)*radius_football))) picture	// football
	= picture
where
	{w,h}					= rectangleSize newFrame
	{fwidth,flength}		= theField
	ratio					= min ((toReal w) / flength) ((toReal h) / fwidth)
	pix x					= toInt (ratio * x)				// convert metres   to pixels
	ppix {px,py}			= {x =pix px,y =pix py}			// convert Position to pixels
	vpix {dx,dy}			= {vx=pix dx,vy=pix dy}			// convert RVector  to pixels
	(northPole,southPole)	= goal_poles theField
	theBall					= getFootball match.theBall (team1 ++ team2)
	
	drawPlayer				:: !Colour !Footballer !*Picture -> *Picture
	drawPlayer colour fb=:{name,nose=nose_dir,effect,pos} picture
		# picture			= drawName          name picture
		| perhaps isOnTheGround effect
							= drawLyingPlayer    fb  picture
		| otherwise			= drawStandingPlayer fb  picture
	where
		player_width		= 0.5
		body_circle			= circle (pix player_width)
		bearing_circle		= circle (pix 0.3)
		nose_dir_vector		= {dx = cos nose_dir, dy = sin nose_dir }
		offset				= {vx = -1, vy = -1}
		nose_dir_point k	= move_point (scale_vector k nose_dir_vector) pos
		
		drawName name picture
			# picture		= setPenColour White                                  picture
			# picture		= drawAt (ppix (move_point {dx=1.0,dy=1.0} pos)) name picture
			= picture
		
		drawLyingPlayer fb=:{length} picture
			# (w,picture)	= getPenSize                                                 picture
			# picture		= setPenSize (pix player_width+2)                            picture
			# picture		= setPenColour Black                                         picture
			# picture		= draw (body_line offset)                                    picture
			# picture		= setPenSize (pix player_width)                              picture
			# picture		= setPenColour colour                                        picture
			# picture		= draw (body_line zero)                                      picture
			# picture		= setPenSize w                                               picture
			# picture		= fillAt (ppix (nose_dir_point (length+0.2))) bearing_circle picture
			# picture		= setPenColour Black                                         picture
			# picture		= drawAt (ppix (nose_dir_point (length+0.2))) bearing_circle picture
			= picture
		where
			body_line v		= { line_end1 = movePoint v (ppix pos)
							  , line_end2 = movePoint v (ppix (nose_dir_point length)) 
							  }
		
		drawStandingPlayer fb picture
			# picture		= setPenColour colour                               picture
			# picture		= fillAt (ppix pos) body_circle                     picture				// body
			# picture		= fillAt (ppix (nose_dir_point 1.0)) bearing_circle picture				// nose direction
			# picture		= setPenColour Black                                picture				// surrounding circle
			# picture		= drawAt (ppix pos) body_circle                     picture
			# picture		= drawAt (ppix (nose_dir_point 1.0)) bearing_circle picture				// nose direction
			= picture
			
	drawMiddleArea			:: !*Picture -> *Picture
	drawMiddleArea picture
		# picture			= drawAt (ppix {center & py=0.0}) v_vector                 picture			// middle line
		# picture			= drawAt (ppix center) (circle (pix radius_centre_circle)) picture			// middle area
		# picture			= drawAt (ppix center) (circle (pix radius_centre_spot))   picture			// center spot
		= picture
	where
		v_vector			= vpix {zero & dy=fwidth}
		center				= {px=flength/2.0, py=fwidth/2.0}
	
	drawGoals				:: !*Picture -> *Picture
	drawGoals picture
		# picture			= drawAt (ppix {px=flength-goal_indent,py=northPole}) h_vector picture		// east north pole
		# picture			= drawAt (ppix {px=flength-goal_indent,py=southPole}) h_vector picture		// east south pole
		# picture			= drawAt (ppix {px=0.0,                py=northPole}) h_vector picture		// west north pole
		# picture			= drawAt (ppix {px=0.0,                py=southPole}) h_vector picture		// west south pole
		= picture
	where
		goal_indent			= 1.0
		h_vector			= vpix {zero & dx=goal_indent}
	
	drawGoalAreas			:: !*Picture -> *Picture
	drawGoalAreas picture
		# picture			= drawAt (ppix {px=flength-goal_area_depth,py=northPole-goal_area_depth}) h_vector picture			// east north edge
		# picture			= drawAt (ppix {px=flength-goal_area_depth,py=southPole+goal_area_depth}) h_vector picture			// east south edge
		# picture			= drawAt (ppix {px=flength-goal_area_depth,py=northPole-goal_area_depth}) v_vector picture			// east edge
		# picture			= drawAt (ppix {px=0.0,                    py=northPole-goal_area_depth}) h_vector picture			// west north edge
		# picture			= drawAt (ppix {px=0.0,                    py=southPole+goal_area_depth}) h_vector picture			// west south edge
		# picture			= drawAt (ppix {px=goal_area_depth,        py=northPole-goal_area_depth}) v_vector picture			// west edge
		= picture
	where
		h_vector			= vpix {zero & dx=goal_area_depth}
		v_vector			= vpix {zero & dy=2.0*goal_area_depth + goal_width}
	
	drawPenaltyAreas		:: !*Picture -> *Picture
	drawPenaltyAreas picture
		# picture			= drawAt (ppix {px=flength-penalty_area_depth,py=northPole-penalty_area_depth}) h_vector picture	// east north edge
		# picture			= drawAt (ppix {px=flength-penalty_area_depth,py=southPole+penalty_area_depth}) h_vector picture	// east south edge
		# picture			= drawAt (ppix {px=flength-penalty_area_depth,py=northPole-penalty_area_depth}) v_vector picture	// east edge
		# picture			= drawAt (ppix east_spot) (circle (pix radius_penalty_spot))                             picture	// east penalty spot
		# picture			= drawAt (ppix (move_point {dx= ~h_dist,dy= ~(sin angle * radius_penalty_area)} east_spot))			// east curve
							         {curve_oval      = circle (pix radius_penalty_area)
							         ,curve_from      = pi-angle
							         ,curve_to        = pi+angle
							         ,curve_clockwise = False }                                                      picture
		# picture			= drawAt (ppix {px=0.0,               py=northPole-penalty_area_depth}) h_vector         picture	// west north edge
		# picture			= drawAt (ppix {px=0.0,               py=southPole+penalty_area_depth}) h_vector         picture	// west south edge
		# picture			= drawAt (ppix {px=penalty_area_depth,py=northPole-penalty_area_depth}) v_vector         picture	// west edge
		# picture			= drawAt (ppix west_spot) (circle (pix radius_penalty_spot))                             picture	// west penalty spot
		# picture			= drawAt (ppix (move_point {dx=h_dist,dy= ~(sin angle * radius_penalty_area)} west_spot))			// west curve
							         {curve_oval      = circle (pix radius_penalty_area)
							         ,curve_from      = angle
							         ,curve_to        = ~angle
							         ,curve_clockwise = True }                                                       picture
		= picture
	where
		h_vector			= vpix {zero & dx=penalty_area_depth}
		v_vector			= vpix {zero & dy=2.0*penalty_area_depth + goal_width}
		angle				= acos (h_dist/radius_penalty_area)
		h_dist				= penalty_area_depth - penalty_spot_depth
		east_spot			= {px=flength-penalty_spot_depth,py=fwidth/2.0}
		west_spot			= {px=        penalty_spot_depth,py=fwidth/2.0}
	
	drawCornerAreas			:: !*Picture -> *Picture
	drawCornerAreas picture
		# picture			= drawAt (ppix {px=        radius_corner_kick_area,py=0.0})    (corner 0.0 (1.5*pi) True ) picture	// west north corner
		# picture			= drawAt (ppix {px=        radius_corner_kick_area,py=fwidth}) (corner 0.0 (0.5*pi) False) picture	// west south corner
		# picture			= drawAt (ppix {px=flength-radius_corner_kick_area,py=0.0})    (corner pi  (1.5*pi) False) picture	// east north corner
		# picture			= drawAt (ppix {px=flength-radius_corner_kick_area,py=fwidth}) (corner pi  (0.5*pi) True ) picture	// east south corner
		= picture
	where
		corner f t c		= {curve_oval=circle (pix radius_corner_kick_area),curve_from=f,curve_to=t,curve_clockwise=c}


green						:== RGB {r=30,g=140,b=40}			// dark green works better on a beamer
home_colour					:== RGB {r=255,g=255,b=70}			// yellow works well on dark green background
away_colour					:== RGB {r=102,g=255,b=255}
ball_colour					:== Black

circle r					:== {oval_rx=r,oval_ry=r}

WestColour					:: Colour
WestColour					= home_colour
EastColour					:: Colour
EastColour					= away_colour

/**	renderAttributes
		yields the attributes required by the rendering functions to accurately draw a match.
*/
renderAttributes ::			!*env -> (![PenAttribute],!*env) | accScreenPicture env
renderAttributes env		= accScreenPicture attributes env
where
	attributes ::			!*Picture -> (![PenAttribute],!*Picture)
	attributes picture
		# ((_,font),picture)= openFont SmallFontDef picture
		= ([PenBack green,PenFont font],picture)
