implementation module Ivanov

import StdEnvExt
import matchControl

chanceOfPenaltySuccess	:== 0.3 
chanceOfSchwalbeSuccess	:== 0.7
chanceOfTheaterSuccess	:== 0.8
chanceOfCatchSuccess	:== 0.5
healthInaccurateFactor	:== 1.0 // (is multiplied with 0.5)
nearEventRadius			:== 10.0
replaceDistance			:== 5.0

ivanovPics :: [Path]
ivanovPics		= map inDir [ "Ivanov_warning"
	                        , "ivanov_yellow"
	                        , "ivanov_red"
	                        , "ivanov_badluck"
	                        , "ivanov_wijst_links"
	                        , "ivanov_wijst_rechts"
	                        , "ivanov_badluck"
	                        , "ivanov_fluit"
	                        , "ivanov_fluit"
	                        , "ivanov_theater"
	                        ]
where
	inDir img	= "afbeeldingen\\"<+++img<+++".bmp"

ivanovReferee :: !FootballField -> Referee
ivanovReferee field = { name  = "Ivanov"
				      , brain = { m   = mkIvanovLongTermMemory
				                , ai  = ref_brain_ivanov field
				                }
				      , refActionPics = ivanovPics
				      }

//	Memory that is passed around for the referee
::	IvanovLongTermMemory =	{ lastRoundTackles			:: ![TackleAction]
							, inOffsidePosition			:: ![FootballerID]
							, keeper1HadBall			:: !Bool							// True iff keeper of Team1 had ball in previous round
							, keeper2HadBall			:: !Bool							// True iff keeper of Team2 had ball in previous round
							, prevHealthPlayers			:: !AssocList FootballerID Health	// the health of the players in previous round
							, lastKickedTheBall			:: !Maybe FootballerID
							, ballIsFor					:: !Maybe ATeam
							, offsidePossible			:: !FreeKickCountdownForOffside // because of a certain type of free kick
							, typeOfKickoff				:: !Maybe RefereeAction
							, waitingForSideSkipping	:: !(!Bool,!Int)
							, firstKick					:: !Maybe ATeam
							, gameLength				:: !Maybe Real
							, receivedYellow			:: ![FootballerID]
							}
::	TackleAction		:== (!Victim,!Offender,!Velocity)
::	Victim				:== (!FootballerID,!Position,!Health)	// Victim   of a tackle
::	Offender			:== (!FootballerID,!Position,!Health)	// Offender of a tackle

/*	KickForcedByReferee : The next kick should be a kick that is appointed by the referee and therefor can not be offside
	FreeToKick :          The first gain or kick after a free kick; can not be offside
	OffsidePossible :     It is possible that you are standing offside
*/
::	FreeKickCountdownForOffside
	= OffsidePossible
	| FreeToKick
	| KickForcedByReferee
instance == FreeKickCountdownForOffside where 
	(==) OffsidePossible     OffsidePossible		= True
	(==) FreeToKick          FreeToKick				= True
	(==) KickForcedByReferee KickForcedByReferee	= True
	(==) _                   _						= False

lowerOffsideCounter :: !IvanovLongTermMemory -> IvanovLongTermMemory
lowerOffsideCounter longMem
	= {longMem & typeOfKickoff   = Nothing
	           , offsidePossible = if (longMem.offsidePossible == KickForcedByReferee) FreeToKick OffsidePossible
	  }

//	Memory used within one timeslice. Is discarded every round.
::	IvanovShortTermMemory =	{ gameStoppedForTackle	:: !Bool
							, punished				:: ![PunishedPlayer]
							, penalty				:: ![(ATeam,Reason)]
							, thisRoundTackles		:: ![TackleAction]
							, myMood				:: !Mood
							, offsideAndTouchedBall	:: !Maybe FootballerID
							, ballKickedOrHeaded	:: !Bool
							}
::	PunishedPlayer		:== (!FootballerID,!PunishedScore,![Reason])
::	PunishedScore		:== Int 	// < 0 points is probably nothing
									// 1-2 points is probably warning
									// 3-4 points is probably yellow
									// > 4 points is probably red

//	Reason of punishments
::	Reason				= R_Tackle 
						| R_Offside 
						| R_Theater 
						| R_Schwalbe 
						| R_Hands 
						| R_DangerousPlay 
						| R_IllegalBallPossession
instance ==  Reason where 
	(==) R_Tackle                R_Tackle					= True
	(==) R_Offside               R_Offside					= True
	(==) R_Theater               R_Theater					= True
	(==) R_Schwalbe              R_Schwalbe					= True
	(==) R_Hands                 R_Hands					= True
	(==) R_DangerousPlay         R_DangerousPlay			= True
	(==) R_IllegalBallPossession R_IllegalBallPossession	= True
	(==) _                       _							= False

mkIvanovShortTermMemory :: !Mood -> IvanovShortTermMemory
mkIvanovShortTermMemory m = { gameStoppedForTackle  = False
							, punished              = []
							, penalty               = []
							, thisRoundTackles      = []
							, myMood                = m
							, offsideAndTouchedBall = Nothing
							, ballKickedOrHeaded    = False
							}

::	Mood		:== P			// Random factor used to randomly pick action when Ivanov is not sure about what happened


mkIvanovLongTermMemory :: IvanovLongTermMemory
mkIvanovLongTermMemory = { lastRoundTackles			= []
						 , inOffsidePosition		= []
						 , keeper1HadBall			= False
						 , keeper2HadBall			= False
						 , prevHealthPlayers		= []
						 , lastKickedTheBall		= Nothing
						 , ballIsFor				= Nothing
						 , offsidePossible			= OffsidePossible
						 , typeOfKickoff			= Nothing
						 , waitingForSideSkipping	= (False, zero)
						 , firstKick				= Nothing
						 , gameLength				= Nothing
						 , receivedYellow			= []
						 }
	
//	Has #param2 done something naughty? In other words, is #param2 known in the list provided as #param1?
getNaughtyFootballer :: ![PunishedPlayer] !FootballerID -> Maybe PunishedPlayer
getNaughtyFootballer punished fbID`
	= case [pun \\ pun=:(fbID,_,_) <- punished | fbID == fbID`] of
		[pun:_]		= Just pun
		_			= Nothing

//	Punish footballer, if he was already punished before, punish him more, otherwise just punish him. Do not make duplicates.
punishMore :: !Reason !Footballer !PunishedScore !IvanovShortTermMemory -> IvanovShortTermMemory
punishMore reason fb score shortMem
	= {shortMem & punished = new_punished}
where
	new_punished	= case break (\(fbID,_,_) -> identify_player fbID fb) shortMem.punished of
						(before,[p],after)	= unbreak (before,[addPunishScore reason score p],after)
						(before,[],[])		= before ++ [(fb.playerID,score,[reason])]
						otherwise			= abort "punishMore: short term memory (punished) contains duplicate entries.\n"

//	Give the footballer a higher punishscore and add the reason if he was not punished before for the same reason	
addPunishScore :: !Reason !PunishedScore !PunishedPlayer -> PunishedPlayer
addPunishScore reason new_score (fbID,score,reasons)
	= (fbID,score+new_score,removeDup (reasons ++ [reason]))

successfulActions :: !Team -> [FootballerWithEffect]
successfulActions players				= map (\{effect,playerID} -> (effect,playerID)) players

//	How this referee thinks about the match and come to his conclusions.
ref_brain_ivanov :: !FootballField !PlayingTime !TimeUnit !FootballState !Half !Team !Team !(!IvanovLongTermMemory,!RandomSeed)
                                                                     -> (![RefereeAction], !(!IvanovLongTermMemory,!RandomSeed))
ref_brain_ivanov field time dt ballState half team1 team2 (longTermMemory,seed)
	# longTermMemory					= if (isJust longTermMemory.gameLength) longTermMemory {longTermMemory & gameLength = Just time}
	// filter actions (what mean actions do I notice and what not)
	# (team1Actions,seed)				= filterOutMeanActions allPlayers (successfulActions team1) seed
	// filter actions team2
	# (team2Actions,seed)				= filterOutMeanActions allPlayers (successfulActions team2) seed
	// look into the consequences of last round tackles
	# (p,seed)							= nextRandomP seed
	# (shortTermMemory, seed)			= analyseTackles (mkIvanovShortTermMemory p) team1Actions team2Actions longTermMemory.lastRoundTackles seed
	// analyse the taken actions
	# (shortTermMemory,longTermMemory)	= analyseActions Team1 team1Actions shortTermMemory longTermMemory
	# (shortTermMemory,longTermMemory)	= analyseActions Team2 team2Actions shortTermMemory longTermMemory
	
	// check for other people who are on the ground (for some mysterious reason)
	# shortTermMemory					= analysePeopleOnTheGround longTermMemory.receivedYellow shortTermMemory team1 team2
	
	// give end-verdict of this round
	# (conclusions,longTermMemory)		= drawConclusions shortTermMemory longTermMemory
	= (conclusions,(longTermMemory,seed))
where
	allPlayers							= team1 ++ team2
	theBall								= getFootball ballState allPlayers
	
	// though teams are returned, only positions AND directions will be altered by a Displace-action
	replacePlayers :: !Team !Team !RefereeAction !FootballField !Football !IvanovLongTermMemory -> Displacements
	replacePlayers team1 team2 kickAction field theBall longTermMemory
		| isForKeeper kickAction Team1				= replaceFielders team1 pos ++ replaceTeam     team2 pos
		| isForKeeper kickAction Team2				= replaceTeam     team1 pos ++ replaceFielders team2 pos
		| isCenterKick kickAction
			# kickOffTeam1							= [(playerID,pos) \\ {playerID,pos} <- getTeam (nameOf team1) (apply field (apply West allAvailableTeams))]
			# kickOffTeam2							= [(playerID,pos) \\ {playerID,pos} <- getTeam (nameOf team2) (apply field (apply East allAvailableTeams))]
			| half == FirstHalf						= kickOffTeam1 ++ kickOffTeam2
			| otherwise								= map (\(fbID,pos) -> (fbID,mirror field pos)) (kickOffTeam1 ++ kickOffTeam2)
		| isForTeam kickAction Team1				= replaceExceptPlayerCloseToPos team1 pos ++ replaceTeam team2 pos
		| otherwise									= replaceTeam team1 pos ++ replaceExceptPlayerCloseToPos team2 pos
	where	
		pos											= case getKickPos field half kickAction of
														Just p	= p
														nothing	= theBall.ballPos.pxy
		
		replaceTeam :: !Team !Position -> Displacements
		replaceTeam team pos						= replacePlayers` pos replaceDistance team
		
		replaceFielders :: !Team !Position -> Displacements
		replaceFielders team pos					= replacePlayers` pos replaceDistance (filter isFielder team)
				
		replaceExceptPlayerCloseToPos :: !Team !Position -> Displacements
		replaceExceptPlayerCloseToPos team pos		= replacePlayers` pos replaceDistance (filter (not o (identify_player closest_player.playerID)) team)
		where
			closest_player							= getClosestPlayer team pos (nameOf team)
	
		isForKeeper :: !RefereeAction !ATeam -> Bool
		isForKeeper (GoalKick team) theTeam			= team==theTeam
		isForKeeper _ _								= False
		
		isForTeam :: !RefereeAction !ATeam -> Bool
		isForTeam (DirectFreeKick team1 _) team2	= team1==team2
		isForTeam (GoalKick       team1)   team2	= team1==team2
		isForTeam (Corner         team1 _) team2	= team1==team2
		isForTeam (ThrowIn        team1 _) team2	= team1==team2
		isForTeam (Penalty        team1)   team2	= team1==team2
		isForTeam (CenterKick     team1)   team2	= team1==team2
		isForTeam _ _								= True
		
		getClosestPlayer :: ![Footballer] !Position !String -> Footballer
		getClosestPlayer []  pos error				= abort ("getClosestPlayer: no player to pick from " <+++ error)
		getClosestPlayer fbs pos error				= winner (isCloser pos) fbs
		where
			isCloser :: !Position !Footballer !Footballer -> Footballer
			isCloser pos fb1=:{pos=pos1} fb2=:{pos=pos2}
				| dist pos pos1 <= dist pos pos2	= fb1
				| otherwise							= fb2			
			
		replacePlayers` :: !Position !Real ![Footballer] -> Displacements
		replacePlayers` pos radius players			= [displace (dist fb.pos pos) fb \\ fb <- players | dist fb.pos pos < radius]
		where
			displace dist fb						= (fb.playerID,new_pos)
			where
				new_pos								= if (dist <= 0.1*radius)
													     (alterPos (move_point (scale_vector radius {dx=cos (dist*2.0*pi),dy=sin (dist*2.0*pi)}) pos))
													     (alterPos (move_point (scale_vector radius {dx=cos angle,        dy=sin angle})         pos))
				angle								= atan ((fb.pos.py - pos.py) / (fb.pos.px - pos.px))
			
		//	corrects when players are placed beyond the borders of the field
			alterPos :: !Position -> Position
			alterPos mypos 
				| mypos.px >= field.flength			= alterPos` (Just (field.flength - one)) Nothing mypos pos
				| mypos.px <= zero					= alterPos` (Just one) Nothing mypos pos
				| mypos.py >= field.fwidth			= alterPos` Nothing (Just (field.fwidth - one)) mypos pos
				| mypos.py <= zero					= alterPos` Nothing (Just one) mypos pos
				| otherwise							= mypos
			where
				alterPos` :: !(Maybe XPos) !(Maybe YPos) !Position !Position -> Position
				alterPos` xpos ypos myPos middlePos 
					| isJust xpos && isJust ypos	= { px = fromJust xpos, py = fromJust ypos }
					| isJust xpos					= { px = fromJust xpos
													  , py = if (myPos.py > middlePos.py)
													            (if (myPos.py < field.fwidth - amount)
													                (myPos.py + amount) 
													                (myPos.py - amount - 2.0*dist_y)
													            ) 
													            (if (myPos.py > amount)
													                (myPos.py - amount) 
													                (myPos.py + amount + 2.0*dist_y)
													            )
													  }
					with	amount					= abs (fromJust xpos - myPos.px)
					| isJust ypos					= { py = fromJust ypos
													  , px = if (myPos.px > middlePos.px)
													            (if (myPos.px + amount < field.flength)
													                (myPos.px + amount)
													                (myPos.px - amount - 2.0*dist_x)
													            )
													            (if (myPos.px > amount)
													                (myPos.px - amount)
													                (myPos.px + amount + 2.0*dist_x)
													            )
													  }
					with	amount					= abs (fromJust ypos - myPos.py)
					| otherwise						= myPos
				where
					dist_y							= abs (myPos.py - middlePos.py)
					dist_x							= abs (myPos.px - middlePos.px)
	
	analyseTackles :: !IvanovShortTermMemory ![FootballerWithEffect] ![FootballerWithEffect] ![TackleAction] !RandomSeed -> (!IvanovShortTermMemory,!RandomSeed)
	//	were there any tackles in the last round?? NO
	analyseTackles shortTermMemory _ _ [] seed
		= (shortTermMemory,seed)
	//	YES: there were tackles last round, whistle for that and ignore all other irrelative events
	analyseTackles shortTermMemory team1Actions team2Actions 
	               [((victim=:{playerNr=v_nmbr},victimPos,victimHealth), (offender=:{clubName=o_cn,playerNr=o_nmbr},offenderPos,offenderHealth), velo) : _] seed	// PA: hmmm, tl of tackles are ignored here.
		# offendersTeam					= if (nameOf team1 == o_cn) Team1 Team2
		# victimNow						= find1 (identify_player victim)   allPlayers
		# offenderNow					= find1 (identify_player offender) allPlayers
		# shortTermMemory				= {shortTermMemory & gameStoppedForTackle = True}
	//	was the ball near the tackle event(s)
		# punishScore					= if (dist theBall.ballPos.pxy offenderPos <= 6.0) 2 1				// ball was near tackle event(s)
										    +
										  if (v_nmbr==1) 5 (if (velo >= 6.0) 3 2)							// penalty for keeper victim is highest
	//	victim is playing theater?
		# actionOfVictim				= case flatten [maybeToList fa \\ (fa,fb_id) <- if (offendersTeam==Team1) team2Actions team1Actions | identify_player fb_id victimNow] of
											[fa:_]	= Just fa
											_		= Nothing
	//	what is the damage to the victim?
		# healthDrop					= victimHealth - victimNow.health + (p-0.5)* healthInaccurateFactor
	//	is the victim making theater?
		# victimPunishScore				= if (isJust actionOfVictim) (if (isPlayedTheater (fromJust actionOfVictim)) 3 0) 0
	//	victim is not playing theater?
		# punishScore					= if (victimPunishScore == 0)
											(definePunishmentOnHealthDrop healthDrop punishScore)
											punishScore
	//	was the ball on 20% part of the field to the offender's own goal?
		# ballWasNearGoal				=  half == FirstHalf  && o_cn == nameOf team1 && (theBall.ballPos.pxy.px >= 0.8 * field.flength || theBall.ballPos.pxy.px <= 0.2 * field.flength)
										|| half == SecondHalf && o_cn == nameOf team2 && (theBall.ballPos.pxy.px >= 0.8 * field.flength || theBall.ballPos.pxy.px <= 0.2 * field.flength)
		# ballWasInPenaltyArea			=  o_cn == nameOf team1 && inPenaltyArea field (opponentHome Team2 half) theBall.ballPos.pxy
										|| o_cn == nameOf team2 && inPenaltyArea field (opponentHome Team1 half) theBall.ballPos.pxy
		# punishScore					= punishScore + if ballWasNearGoal 1 0
		# shortTermMemory				= if ballWasInPenaltyArea
											{shortTermMemory & penalty = shortTermMemory.penalty ++ [(offendersTeam,R_Tackle)]}
											shortTermMemory
		# offenderFromSMemory			= getNaughtyFootballer shortTermMemory.punished offender
		# victimFromSMemory				= getNaughtyFootballer shortTermMemory.punished victim
		# offenderVerdict				= if (isJust offenderFromSMemory)
											(addPunishScore R_Tackle punishScore (fromJust offenderFromSMemory))
											(offender,punishScore,[R_Tackle])
		# victimVerdict					= if (isJust victimFromSMemory)
											(addPunishScore R_Theater victimPunishScore (fromJust victimFromSMemory))
											(victim,victimPunishScore,[R_Theater]) 
		= if (snd3 offenderVerdict == 0)
			if (snd3 victimVerdict == 0)
				(shortTermMemory,seed1)
				({shortTermMemory & punished = shortTermMemory.punished ++ [victimVerdict]},seed1)
			if (snd3 victimVerdict == 0)
				({shortTermMemory & punished = shortTermMemory.punished ++ [offenderVerdict]},seed1)
				({shortTermMemory & punished = shortTermMemory.punished ++ [offenderVerdict,victimVerdict]},seed1)		
	where
		(p,seed1)						= nextRandomP seed
		
		definePunishmentOnHealthDrop :: !Real !Int -> Int
		definePunishmentOnHealthDrop healthDrop punishScore
			| healthDrop < 0.5			= punishScore
			| 0.5 <= healthDrop && healthDrop <= 1.5
										= punishScore + 1
			| 1.5 < healthDrop && healthDrop <= 3.5
										= punishScore + 2
			| otherwise					= punishScore + 3
	
	analyseActions :: !ATeam ![FootballerWithEffect] !IvanovShortTermMemory !IvanovLongTermMemory
	                                             -> (!IvanovShortTermMemory,!IvanovLongTermMemory)
	analyseActions team [] shortTermMemory longTermMemory
		= (shortTermMemory,{longTermMemory & lastRoundTackles = []})
	analyseActions team [action=:(fa,{clubName=cn,playerNr=nmbr}):actions] shortTermMemory longTermMemory
	//	Am I in the pause between the first and the second half (long enough to allow switching of sides)
		| fst longTermMemory.waitingForSideSkipping
			| snd longTermMemory.waitingForSideSkipping <= zero
				= (shortTermMemory,{longTermMemory & waitingForSideSkipping = (False,zero)})
			| otherwise
				= (shortTermMemory,{longTermMemory & waitingForSideSkipping = (True,snd longTermMemory.waitingForSideSkipping-one)})
		# (team1Name,currHome)	= if (half==FirstHalf) (nameOf team1,West) (nameOf team1,East)
	//	did I stop the game because of tackle-actions?
		| shortTermMemory.gameStoppedForTackle
		//	No more analysing game, but still look at mean actions with a connection to the tackle(s)
			| perhaps isSchwalbed fa
				|  isTackleVictim   taf.playerID longTermMemory.lastRoundTackles	// was he victim
				|| isTackleOffender tef.playerID longTermMemory.lastRoundTackles	// or offender
				|| nearLastRoundTackle fb.pos longTermMemory.lastRoundTackles		// or near a tackle?
					| inPenaltyArea field (opponentHome team half) fb.pos			// was it in penaltyarea?
						= analyseActions team actions {shortTermMemory & penalty = shortTermMemory.penalty ++ [(team,R_Schwalbe)]} longTermMemory
					| otherwise														// punish him or punish him more
						= analyseActions team actions (punishMore R_Schwalbe tef 1 shortTermMemory) longTermMemory
				| otherwise															// otherwise ignore
					= analyseActions team actions shortTermMemory longTermMemory//ignoreThisAction
			| perhaps isPlayedTheater fa
				|  isTackleOffender tef.playerID longTermMemory.lastRoundTackles	// was he offender
				|| nearLastRoundTackle fb.pos longTermMemory.lastRoundTackles		// or near a tackle
				&& not (isTackleVictim taf.playerID longTermMemory.lastRoundTackles)// and is not a victim
					| inPenaltyArea field (opponentHome team half) fb.pos			// was it in penaltyarea?
						= analyseActions team actions {shortTermMemory & penalty = shortTermMemory.penalty ++ [(team,R_Theater)]} longTermMemory
					| otherwise
						= analyseActions team actions (punishMore R_Theater tef 1 shortTermMemory) longTermMemory
			//	victims are already scanned during tackle analyses to see if it was worth to check the health of the victim
				| otherwise															// otherwise ignore
					= analyseActions team actions shortTermMemory longTermMemory//ignoreThisAction
			| perhaps isTackled fa
				|  isTackleVictim   taf.playerID (longTermMemory.lastRoundTackles)	// was he victim
				|| isTackleOffender tef.playerID (longTermMemory.lastRoundTackles)	// or offender
				|| nearLastRoundTackle fb.pos longTermMemory.lastRoundTackles		// or near a tackle?
					| inPenaltyArea field (opponentHome team half) fb.pos			// was it in penaltyarea?
						= analyseActions team actions {shortTermMemory & penalty = shortTermMemory.penalty ++ [(team,R_Schwalbe)]} longTermMemory
					| otherwise														// punish him or punish him more
						= analyseActions team actions (punishMore R_Tackle tef 2 shortTermMemory) longTermMemory
				//otherwise ignore
			//	Usually ignore, but: a tackle is something serious... when I'm in a bad mood, I will not ignore this one
				| shortTermMemory.myMood > 0.8
					| inPenaltyArea field (opponentHome team half) fb.pos			// was it in penaltyarea?
						= analyseActions team actions {shortTermMemory & penalty = shortTermMemory.penalty ++ [(team,R_Schwalbe)]} longTermMemory
					| otherwise
						= analyseActions team actions (punishMore R_Tackle tef 3 shortTermMemory) longTermMemory
				| otherwise															// player was lucky...
					= analyseActions team actions shortTermMemory longTermMemory//ignoreThisAction
			| perhaps isCaughtBall fa
				| isKeeper tef && inKeeperSpot tef team1Name currHome field			// is he a keeper?
					= analyseActions team actions shortTermMemory longTermMemory//ignoreThisAction
				|  isTackleVictim   taf.playerID longTermMemory.lastRoundTackles	// was he victim
				|| isTackleOffender tef.playerID longTermMemory.lastRoundTackles	// or offender
				|| nearLastRoundTackle fb.pos longTermMemory.lastRoundTackles		// or near a tackle?
					| inPenaltyArea field (opponentHome team half) fb.pos			// was it in penaltyarea?
						= analyseActions team actions {shortTermMemory & penalty = shortTermMemory.penalty ++ [(team,R_Schwalbe)]} longTermMemory
					| otherwise														// punish him or punish him more
						= analyseActions team actions (punishMore R_Hands tef 2 shortTermMemory) longTermMemory
				| otherwise															// otheriwse ignore
					= analyseActions team actions shortTermMemory longTermMemory//ignoreThisAction
		//	All other kind of actions are ignored
			| otherwise
				= analyseActions team actions shortTermMemory longTermMemory//ignoreThisAction
	//	There were no tackle-actions last round -> Ivanov needs to keep thinking
	//	Everyone is free to move over the field
		| perhaps isMoved fa || perhaps isFeinted fa
			= analyseActions team actions shortTermMemory longTermMemory//ignoreThisAction
	//	Gainball is illegal when in offside position
	//	A legal gainBall causes no one to be in offside position (until a kick- or headball event)
		| perhaps isGainedBall fa
			# longTermMemory		= if (isNothing longTermMemory.firstKick) {longTermMemory & firstKick = Just team} longTermMemory		
		//	He was not allowed to get the ball (because of referee-action in the previous round)
			| not (wasAllowedToGetBall longTermMemory.ballIsFor (nameOf team1) (nameOf team2) cn)
				= analyseActions team actions (punishMore R_IllegalBallPossession tef 1 shortTermMemory) longTermMemory
			# longTermMemory		= {longTermMemory & ballIsFor = Nothing}
		//	He was in offside position
			| isMember tef.playerID longTermMemory.inOffsidePosition && longTermMemory.offsidePossible == OffsidePossible 
			//	Detect offside for wistle
				= analyseActions team actions {punishMore R_Offside tef 0 shortTermMemory & offsideAndTouchedBall=Just tef.playerID} longTermMemory
			| otherwise															// clear offsidelist: no one in offside position now
				= analyseActions team actions shortTermMemory {longTermMemory & inOffsidePosition = []}
	//	{Kick/Head}Ball is illegal when in offside position (ball was not gained)
		| perhaps isKickedBall fa || perhaps isHeadedBall fa
			# longTermMemory		= if (isNothing longTermMemory.firstKick) {longTermMemory & firstKick = Just team} longTermMemory	
			# shortTermMemory		= {shortTermMemory & ballKickedOrHeaded = True}
		//	He was not allowed to get the ball
			| not (wasAllowedToGetBall longTermMemory.ballIsFor (nameOf team1) (nameOf team2) cn)
				= analyseActions team actions (punishMore R_IllegalBallPossession tef 1 shortTermMemory) longTermMemory
			# longTermMemory		= {longTermMemory & ballIsFor = Nothing}
			| isMember tef.playerID longTermMemory.inOffsidePosition && longTermMemory.offsidePossible == OffsidePossible	// detect offside for wistle
				= analyseActions team actions {punishMore R_Offside tef 0 shortTermMemory & offsideAndTouchedBall=Just tef.playerID} longTermMemory
		//	A legal {Kick/Head}Ball may be illegal/dangerous play when ball was gained by keeper
			# (team1Name,currHome)	= if (half == FirstHalf) (nameOf team1,West) (nameOf team1,East)
			# keeper1HadBall		= let keepers = filter isKeeper team1
									      keeper  = hd keepers
									   in not (isEmpty keepers) && ballIsGainedBy keeper.playerID ballState && inKeeperSpot keeper team1Name currHome field && longTermMemory.keeper1HadBall
			# keeper2HadBall		= let keepers = filter isKeeper team2
									      keeper  = hd keepers
									   in not (isEmpty keepers) && ballIsGainedBy keeper.playerID ballState && inKeeperSpot keeper team1Name currHome field && longTermMemory.keeper2HadBall
			| keeper1HadBall && nmbr<>1 || keeper2HadBall && nmbr<>1
				= analyseActions team actions (punishMore R_DangerousPlay tef 5 shortTermMemory) longTermMemory
		//	A legal {Kick/Head}Ball may put people in a potential offside position
			# longTermMemory				= {longTermMemory & lastKickedTheBall = Just tef.playerID}
			# (homeOfTeam1,homeOfTeam2)		= if (half == FirstHalf) (West,East) (East,West)
			| team == Team1
				# closestDefenderTeam2Xpos	= offsideline field homeOfTeam2 theBall team2
				# inOffsidePos				= getOffsidePlayers field homeOfTeam1 closestDefenderTeam2Xpos (removeMember tef team1)
				= analyseActions team actions shortTermMemory {longTermMemory & inOffsidePosition = inOffsidePos}
			| otherwise
				# closestDefenderTeam1Xpos	= offsideline field homeOfTeam1 theBall team1
				# inOffsidePos				= getOffsidePlayers field homeOfTeam2 closestDefenderTeam1Xpos (removeMember tef team2)
				= analyseActions team actions shortTermMemory {longTermMemory & inOffsidePosition = inOffsidePos}
		| perhaps isTackled fa
		//	Store him, so we can look next round at the damage of the victim
			# (victimID,ve)		= case fromJust fa of (Tackled victimID ve _) -> (victimID,ve)
			# moreTackles		= case filter (identify_player victimID) (team1++team2) of
									[fb:_]	= [((fb.playerID,fb.pos,fb.health),(tef.playerID,tef.pos,tef.health),ve)]
									none	= []
			# shortTermMemory	= {shortTermMemory & thisRoundTackles = shortTermMemory.thisRoundTackles ++ moreTackles}
			= analyseActions team actions shortTermMemory longTermMemory
		| perhaps isSchwalbed fa
			# opponents			= if (team == Team1) team2 team1
			# opponentNear		= not (isEmpty [pos \\ {pos} <- opponents | dist fb.pos pos <= 5.0])
			# ballNear			= dist {zero & pxy=fb.pos} theBall.ballPos <= 10.0
			# shortTermMemory	= if (inPenaltyArea field (opponentHome team half) fb.pos)	// In penaltyarea?
									 {shortTermMemory & penalty = shortTermMemory.penalty ++ [(team,R_Schwalbe)]}
								 (if (opponentNear && ballNear)								// Was an opponent near and was the ball near? Then punish 4
								     (punishMore R_Schwalbe tef 4 shortTermMemory)
								 (if ballNear												// Was the ball near? Then punish 2
								     (punishMore R_Schwalbe tef 2 shortTermMemory)
								 (if opponentNear											// Was an opponent near? Then punish 3
								     (punishMore R_Schwalbe tef 3 shortTermMemory)
								     (punishMore R_Schwalbe tef 1 shortTermMemory)
								 )))
			= analyseActions team actions shortTermMemory longTermMemory
		| perhaps isCaughtBall fa
			# longTermMemory	= if (isNothing longTermMemory.firstKick) {longTermMemory & firstKick = Just team} longTermMemory
			| isKeeper tef && inKeeperSpot tef team1Name currHome field		// is he a keeper (todo: and located in his penaltyarea (for everything, gaining ball, etc.)
				= analyseActions team actions shortTermMemory longTermMemory//ignoreThisAction
			# shortTermMemory	= if (inPenaltyArea field (opponentHome team half) theBall.ballPos.pxy)
								     {punishMore R_Hands tef 2 shortTermMemory & penalty = shortTermMemory.penalty ++ [(team,R_Hands)]}
								     (punishMore R_Hands tef 4 shortTermMemory)
			= analyseActions team actions shortTermMemory longTermMemory
	// theater, but no tackle or schwalbe was seen, so theater will be punished lightly
		| perhaps isPlayedTheater fa
			= analyseActions team actions (punishMore R_Theater tef 1 shortTermMemory) longTermMemory
	// unknown action
		| otherwise
			= analyseActions team actions shortTermMemory longTermMemory//ignoreThisAction
	where
		tef						= find1 (identify_player {clubName=cn,playerNr=nmbr}) allPlayers
		taf						= tef
		fb						= taf
		
		ignoreThisAction		= analyseActions team actions shortTermMemory longTermMemory
		
		wasAllowedToGetBall :: !(Maybe ATeam) !ClubName !ClubName !ClubName -> Bool
		wasAllowedToGetBall Nothing _ _ _			= True
		wasAllowedToGetBall (Just team) team1 team2 myTeam
			| team == Team1		= team1 == myTeam
			| otherwise			= team2 == myTeam		
					
	analysePeopleOnTheGround :: ![FootballerID] !IvanovShortTermMemory !Team !Team -> IvanovShortTermMemory
	analysePeopleOnTheGround receivedYellow shortMem team1 team2
		| shortMem.gameStoppedForTackle		// have I stopped the game already for a tackle
			= shortMem
		| otherwise
			= analysePeopleOnTheGround` receivedYellow shortMem [fb \\ fb <- team1 ++ team2 | lastEventIsOnTheGround fb.effect]
	where
		lastEventIsOnTheGround :: !(Maybe FootballerEffect) -> Bool
		lastEventIsOnTheGround es	= isJust es && isOnTheGround (fromJust es)
		
		analysePeopleOnTheGround` :: ![FootballerID] !IvanovShortTermMemory ![Footballer] -> IvanovShortTermMemory
		analysePeopleOnTheGround` _ shortMem []
											= shortMem
		analysePeopleOnTheGround` receivedYellow shortMem [fb=:{playerID={clubName=cn}}:tefls]
		//	how many opponents could have tackled him
			# suspects						= [s \\ s <- team1 ++ team2 | dist fb.pos s.pos <= maxTackleReach s]
			# healthDrop					= case lookup fb.playerID longTermMemory.prevHealthPlayers of
												Just prevHealth		= fb.health - prevHealth
												none				= zero
			# motive4Schwalbe				= ballDirectionIsTowardsGoal && ballNear
			# goodMotive4Schwalbe			= motive4Schwalbe && ballWasInPenaltyAreaOpponent
			# schwalbeThreshold				= if motive4Schwalbe
												if goodMotive4Schwalbe
													if victimReprimandedBefore (0.5,0.4 ) (0.3,0.25)
													if victimReprimandedBefore (0.2,0.17) (0.1,0.08)
												if victimReprimandedBefore     (0.1,0.08) (-999.0,-999.0)
			# schwalbeThreshold				= if (length suspects > 1)
											     (fst schwalbeThreshold * 0.5, snd schwalbeThreshold * 0.5)
											     schwalbeThreshold
			| not (isEmpty suspects)		// tackle or schwalbe??
				# primarySuspect			= suspects!!((length suspects * (toInt (shortMem.myMood * 10.0))) mod (length suspects) )
				# isNoOtherDefenderLeft		= if (cn == nameOf team1 && half==FirstHalf || cn == nameOf team2 && half==SecondHalf)
												 (isEmpty [fb \\ fb <- filter isFielder team2 | fb.pos.px > primarySuspect.pos.px])
												 if (cn == nameOf team1 && half==SecondHalf || cn == nameOf team2 && half==FirstHalf)
												 	 (isEmpty [fb \\ fb <- filter isFielder team2 | fb.pos.px < primarySuspect.pos.px])
													 (abort "primary suspect of possible tackle does not play in one of the teams")
				# opponentReprimandedBefore	= isMember primarySuspect.playerID receivedYellow // not (isEmpty primarySuspect.events.reprimands) || primarySuspect.events.gotYellow
				# motive4Tackle				= ballDirectionIsTowardsGoal && ballNear || ballGoingTowardsVictim
				# goodMotive4Tackle			= motive4Tackle && (isNoOtherDefenderLeft || goalVeryNear)
				# tackleThreshold			= if motive4Tackle
												if goodMotive4Tackle
													if opponentReprimandedBefore (0.5,0.6 ) (0.7,0.75)
													if opponentReprimandedBefore (0.8,0.83) (0.9,0.92)
												if opponentReprimandedBefore     (0.9,0.92) (999.0,999.0)
				# shortMem					= if (shortMem.myMood >= fst tackleThreshold && shortMem.myMood < snd tackleThreshold && healthDrop > 0.1)	// Give yellow for tackle
											     (punishMore R_Tackle primarySuspect 3 shortMem)
											 (if (shortMem.myMood >= fst tackleThreshold)																// Warn for tackle
											     (punishMore R_Theater primarySuspect 1 shortMem)
											 (if (shortMem.myMood <= fst schwalbeThreshold && shortMem.myMood > snd schwalbeThreshold && healthDrop > 0.1)// Give yellow for schwalbe
											     (punishMore R_Schwalbe fb 3 shortMem)
											 (if (shortMem.myMood <= fst schwalbeThreshold)																// Warn for schwalbe
											     (punishMore R_Schwalbe fb 1 shortMem)
											     shortMem
											 )))
				= analysePeopleOnTheGround` receivedYellow shortMem tefls
			# shortMem						= if (shortMem.myMood <= fst schwalbeThreshold && shortMem.myMood > snd schwalbeThreshold && healthDrop > 0.1)	// Warn good for schwalbe
											     (punishMore R_Schwalbe fb 2 shortMem)
											 (if (shortMem.myMood <= fst schwalbeThreshold)																	// Warn for schwalbe
											     (punishMore R_Schwalbe fb 1 shortMem)
											     shortMem
											 )
			= analysePeopleOnTheGround` receivedYellow shortMem tefls
		where
			goalVeryNear					= if (cn == nameOf team1 && half == FirstHalf  || cn == nameOf team2 && half == SecondHalf)
											     (dist fb.pos {px=field.flength,py=field.fwidth/2.0} <= 20.0)	// east goal
											 (if (cn == nameOf team1 && half == SecondHalf || cn == nameOf team2 && half == FirstHalf)
											     (dist fb.pos {px=zero,py=field.fwidth/2.0} <= 20.0)			// west goal
											     (abort "fallen player is not from one of the teams"))
			ballNear						= dist {zero & pxy=fb.pos} theBall.ballPos <= 8.0
			victimReprimandedBefore			= isMember fb.playerID receivedYellow
			ballWasInPenaltyAreaOpponent	= cn == nameOf team1 && inPenaltyArea field (opponentHome Team1 half) theBall.ballPos.pxy
												||
											  cn == nameOf team2 && inPenaltyArea field (opponentHome Team2 half) theBall.ballPos.pxy
			ballWasInPenaltyAreaSelf		= cn == nameOf team1 && inPenaltyArea field (opponentHome Team2 half) theBall.ballPos.pxy
												||
											  cn == nameOf team2 && inPenaltyArea field (opponentHome Team1 half) theBall.ballPos.pxy
			ballDirectionIsTowardsGoal		= if (cn == nameOf team1 && half == FirstHalf || cn == nameOf team2 && half == SecondHalf)
											     (theBall.ballSpeed.vxy.direction > 0.5*pi && theBall.ballSpeed.vxy.direction < 1.5*pi)	// to east ..
											 (if (cn == nameOf team1 && half == SecondHalf || cn == nameOf team2 && half == FirstHalf)
											     (theBall.ballSpeed.vxy.direction < 0.5*pi || theBall.ballSpeed.vxy.direction > 1.5*pi)	// to west .. PA: odd, why use ..||.. here and ..&&.. immediately above?
											     (abort "fallen player is not from one of the teams"))
			ballGoingTowardsVictim			= nextBallPos theBall.ballPos theBall.ballSpeed (fb.pos,5.0,fb.length)
	
	drawConclusions :: !IvanovShortTermMemory !IvanovLongTermMemory -> (![RefereeAction],!IvanovLongTermMemory)
	drawConclusions shortMem longMem
		# longMem							= {longMem & keeper1HadBall		= ballIsGainedBy {playerNr=1,clubName=nameOf team1} ballState//all hasBall (filter isKeeper team1)
											           , keeper2HadBall		= ballIsGainedBy {playerNr=1,clubName=nameOf team2} ballState//all hasBall (filter isKeeper team2)
											           , prevHealthPlayers	= map (\fb -> (fb.playerID, fb.health)) (team1 ++ team2)
											  }
		# team_penalty						= firstPenalty shortMem.penalty
		# offence_penalty					= getMostImportantKickAction basicActions
		# (kickAction,longMem)				= if (isJust team_penalty)    (let t = fromJust team_penalty    in (Just (Penalty t),{longMem & ballIsFor     = Just t
											                                                                                              , typeOfKickoff = Just (Penalty t)}))
											 (if (isJust offence_penalty) (let p = fromJust offence_penalty in (offence_penalty, {longMem & ballIsFor     = Just (actionForWhichTeam p)
											                                                                                              , typeOfKickoff = offence_penalty}))
											 ((Nothing,longMem)
											 ))
		# noticedActions					= map fst basicActions
		# longMem							= if (isJust kickAction)           {longMem & offsidePossible = KickForcedByReferee}
											 (if (shortMem.ballKickedOrHeaded) (lowerOffsideCounter longMem)
												                               longMem
											 )
		| not shortMem.gameStoppedForTackle		// We haven't analysed tackles this round, it was a normal round
			# refActions					= case kickAction of
												Just kick		= noticedActions ++ [DisplacePlayers (replacePlayers team1 team2 kick field theBall longMem)]
												_				= noticedActions
			| time <= (fromJust longMem.gameLength)/2.0 && half == FirstHalf			// is it time to change half?
				# longMem					= {longMem & waitingForSideSkipping = (True,150)}
				# teamThatMayStartAtSecondHalf
											= if (isJust longMem.firstKick)
												(if ((fromJust longMem.firstKick) == Team1) Team2 Team1)
												Team2
				# centerKick				= CenterKick teamThatMayStartAtSecondHalf
				# ds						= mirror_teams (replacePlayers team1 team2 centerKick field theBall longMem)
				= (refActions ++ [ContinueGame] ++ [EndHalf,centerKick,DisplacePlayers ds],{longMem & ballIsFor = Just teamThatMayStartAtSecondHalf})

			| time <= zero						// is it time to stop the game?
				= (refActions ++ [GameOver],longMem)
		//	Is the ball out of the lines and do we have: a goal, a corner, a goal kick, or a throw in
			# (behindLineActions,longMem) = getBehindLinesActions half theBall field longMem
		//	No matter what happened, ball behind the line is ball behind the line
			| length behindLineActions > zero
				# displacements				= replacePlayers team1 team2 (last behindLineActions) field theBall longMem
				# refActions				= refActions ++ [DisplacePlayers displacements]
				= (refActions ++ [ContinueGame] ++ behindLineActions,{longMem & offsidePossible = KickForcedByReferee})
			# longMem						= {longMem & lastRoundTackles = shortMem.thisRoundTackles}
		//	We have no new tackles for the next round
			| isEmpty shortMem.thisRoundTackles
				| isJust kickAction
					# kick					= fromJust kickAction
					# goalArea				= kickActionIsFreeKickInPenaltyArea kick
					| isJust goalArea
						# displacements		= replacePlayers team1 team2 (fromJust goalArea) field theBall longMem
						= (refActions ++ [ContinueGame,DisplacePlayers displacements,fromJust goalArea],longMem)
					| otherwise					
						= (refActions ++ [ContinueGame,kick],longMem)
				| otherwise
					= (refActions ++ [ContinueGame],longMem)
		//	We have new tackles for the next round
			| otherwise
				= (refActions ++ [PauseGame,AddTime (1.0/30.0)],longMem)
	//	We have analysed tackles this round
		| otherwise
		//	We have a position to restart the game from
			| isJust kickAction
				# kick						= fromJust kickAction
				# goalArea					= kickActionIsFreeKickInPenaltyArea kick
				| isJust goalArea
					# displacements			= replacePlayers team1 team2 (fromJust goalArea) field theBall longMem
					= ([ContinueGame] ++ noticedActions ++ [DisplacePlayers displacements,fromJust goalArea],longMem)
				# displacements				= replacePlayers team1 team2 kick field theBall longMem
				= ([ContinueGame] ++ noticedActions ++ [DisplacePlayers displacements,kick],longMem)
		//	I still have to think about a new place to restart the game from
			| otherwise
				= abort "should have a position to restart the game from"
	where
		basicActions						= getAllBasicActions longMem.receivedYellow shortMem
		
		mirror_teams :: !Displacements -> Displacements
		mirror_teams ds						= map (\(fbID,pos) -> (fbID,mirror field pos)) ds
		
		kickActionIsFreeKickInPenaltyArea :: !RefereeAction -> Maybe RefereeAction
		kickActionIsFreeKickInPenaltyArea (DirectFreeKick team pos)
			| inPenaltyArea field (opponentHome team half) pos
				= Just (Penalty team)
			| inPenaltyArea field (teamHome team half) pos
				= Just (GoalKick team)
			| otherwise
				= Nothing
		kickActionIsFreeKickInPenaltyArea _
			= Nothing
		
		getBehindLinesActions :: !Half !Football !FootballField !IvanovLongTermMemory -> (![RefereeAction],!IvanovLongTermMemory)
		getBehindLinesActions half theBall field longMem=:{lastKickedTheBall}
			| theBall.ballPos.pxy == p			// ball is not behind (see definition of point_to_rectangle)
				= ([],longMem)
			| isNothing lastKickedTheBall		// the ball got behind by unknown forces
				= ([CenterKick Team1],longMem)
			| theBall.ballPos.pxy.px < zero		// behind line West?
				# (team,homeTeamWest)	= if (half == FirstHalf) (Team1,nameOf team1) (Team2,nameOf team2)				// first half?
				| isbetween theBall.ballPos.pxy.py (field.fwidth/2.0 - goal_width/2.0) (field.fwidth/2.0 + goal_width/2.0)	// goal?
					&& 
				  theBall.ballPos.pz < goal_height 
					= if (team==Team1) ([Goal Team2,CenterKick Team1],{longMem & ballIsFor=Just Team1})
					                   ([Goal Team1,CenterKick Team2],{longMem & ballIsFor=Just Team2})
				| (fromJust lastKickedTheBall).clubName == homeTeamWest		// lastPlayerKicked Team1?
					# corner			= if (theBall.ballPos.pxy.py > field.fwidth/2.0) South North
					= ([Corner (other team) corner],{longMem & ballIsFor = Just (other team)})
				| otherwise													// opponent team -> goalkick
					= ([GoalKick team],{longMem & ballIsFor = Just team})
			| theBall.ballPos.pxy.px > field.flength						// behind line East?
				# (team,homeTeamEast)	= if (half == FirstHalf) (Team2,nameOf team2) (Team1,nameOf team1)
				| isbetween theBall.ballPos.pxy.py (field.fwidth/2.0 - goal_width/2.0) (field.fwidth/2.0 + goal_width/2.0)	// goal?
					&& 
				  theBall.ballPos.pz < goal_height 
					= if (team==Team1) ([Goal Team2,CenterKick Team1],{longMem & ballIsFor=Just Team1})
					                   ([Goal Team1,CenterKick Team2],{longMem & ballIsFor=Just Team2})
				| (fromJust lastKickedTheBall).clubName == homeTeamEast		// lastPlayerKicked Team1?
					# corner			= if (theBall.ballPos.pxy.py > field.fwidth/2.0) South North
					= ([Corner (other team) corner],{longMem & ballIsFor = Just (other team)})
				| otherwise
					= ([GoalKick team],{longMem & ballIsFor = Just team})
			# throwin					= if ((fromJust lastKickedTheBall).clubName == nameOf team1) Team2 Team1
			= ([ThrowIn throwin p],{longMem & ballIsFor = Just throwin})
		where
			p							= point_to_rectangle (zero,{px=field.flength,py=field.fwidth}) theBall.ballPos.pxy

	/*	Most important (in given order): Tackle, OwnBallIllegally, Offside, DangerousPlay, Hands 
		Even: the rest
	*/	getMostImportantKickAction :: ![(RefereeAction,RefereeAction)] -> Maybe RefereeAction
		getMostImportantKickAction actions
			| isEmpty actions			= Nothing
			| otherwise					= Just resumeAction
		where
			(_,resumeAction)			= maxListBy (\(a1,r1) (a2,r2) -> priority a1 >= priority a2) actions
			priority action				= if (isTackleDetected   action) 5
										 (if (isOwnBallIllegally action) 4
										 (if (isOffside          action) 3
										 (if (isDangerousPlay    action) 2
										 (if (isHands            action) 1
										                                 0
										 ))))
		
		actionForWhichTeam :: !RefereeAction -> ATeam
		actionForWhichTeam (DirectFreeKick t _)		= t
		actionForWhichTeam (GoalKick       t)		= t
		actionForWhichTeam (Corner         t _)		= t
		actionForWhichTeam (ThrowIn        t _)		= t
		actionForWhichTeam (Penalty        t)		= t
		actionForWhichTeam (CenterKick     t)		= t
		actionForWhichTeam _						= abort "actionForWhichTeam: kick action expected"
			
		getAllBasicActions :: ![FootballerID] !IvanovShortTermMemory -> [(RefereeAction,RefereeAction)]
		getAllBasicActions receivedYellow shortMem		= getAllBasicPunishActions receivedYellow shortMem.punished 
		where
		//	PA: this should really be programmed as a map!
			getAllBasicPunishActions :: ![FootballerID] ![PunishedPlayer] -> [(RefereeAction,RefereeAction)]
			getAllBasicPunishActions _ []		= []
			getAllBasicPunishActions receivedYellow [(fbID=:{clubName=cn},punish,[]):pals]
				| punish > 4
					= [(ReprimandPlayer fbID RedCard,DirectFreeKick otherTeam kickPos) : rest]
				| punish >= 3
					# yellow					= (ReprimandPlayer fbID YellowCard,DirectFreeKick otherTeam kickPos)
					# red						= (ReprimandPlayer fbID RedCard,   DirectFreeKick otherTeam kickPos)
					| isMember fbID receivedYellow
												= [yellow, red : rest]
					| otherwise					= [yellow : rest]
				| punish >= 1					= [(ReprimandPlayer fbID Warning,DirectFreeKick otherTeam kickPos) : rest]
				| otherwise						= rest
			where
				kickPosHome						= if (cn == nameOf team1 && half == FirstHalf || cn == nameOf team2 && half == SecondHalf) East West
				kickPosTeam						= if (cn == nameOf team1) team2 team1
				kickPos							= posForFreeKick kickPosHome theBall.ballPos.pxy kickPosTeam
				otherTeam						= if (cn == nameOf team1) Team2 Team1
				rest							= getAllBasicPunishActions receivedYellow pals
			getAllBasicPunishActions receivedYellow [(fbID=:{clubName=cn},punish,[r:rls]):pals]
				| r == R_Tackle					= [(TackleDetected fbID,DirectFreeKick otherTeam kickPos) : rest]
				| r == R_Offside && isJust shortMem.offsideAndTouchedBall
					| fromJust shortMem.offsideAndTouchedBall == fbID
												= [(Offside          fbID,DirectFreeKick otherTeam kickPos) : rest]
					| otherwise					= rest
				| r == R_Theater				= [(TheaterDetected  fbID,DirectFreeKick otherTeam kickPos) : rest]
				| r == R_Schwalbe				= [(SchwalbeDetected fbID,DirectFreeKick otherTeam kickPos) : rest]
				| r == R_Hands					= [(Hands            fbID,DirectFreeKick otherTeam kickPos) : rest]
				| r == R_DangerousPlay			= [(DangerousPlay    fbID,DirectFreeKick otherTeam kickPos) : rest]
				| r == R_IllegalBallPossession	= [(OwnBallIllegally fbID,DirectFreeKick otherTeam kickPos) : rest]
				| otherwise						= abort "getAllBasicPunishActions: unknown reason of RefereeAction."
			where
				kickPosHome						= if (cn == nameOf team1 && half == FirstHalf || cn == nameOf team2 && half == SecondHalf) East West
				kickPosTeam						= if (cn == nameOf team1) team2 team1
				kickPos							= posForFreeKick kickPosHome theBall.ballPos.pxy kickPosTeam
				otherTeam						= if (cn == nameOf team1) Team2 Team1			
				rest							= getAllBasicPunishActions receivedYellow [(fbID,punish,rls):pals]	
		
		firstPenalty :: ![(ATeam,Reason)] -> Maybe ATeam
		firstPenalty []							= Nothing
		firstPenalty [(t,_):_]					= Just (other t)

isTackleVictim :: !FootballerID ![TackleAction] -> Bool
isTackleVictim fbID tackleEvents				= any (\((fbID`,_,_),_,_) -> fbID == fbID`) tackleEvents

isTackleOffender :: !FootballerID ![TackleAction] -> Bool
isTackleOffender fbID tackleEvents				= any (\(_,(fbID`,_,_),_) -> fbID == fbID`) tackleEvents

nearLastRoundTackle :: !Position ![TackleAction] -> Bool
nearLastRoundTackle pos tackleEvents			= all near tackleEvents
where
	near ((_,vpos,_),(_,opos,_),_)				= dist pos vpos > nearEventRadius && dist pos opos > nearEventRadius
			
//	Used for offside, so we only look at the xposition
offsideline :: !FootballField !Home !Football ![Footballer] -> Metre
offsideline field home ball fbs					= case sortBy compare [ball.ballPos.pxy.px : [fb.pos.px \\ fb <- fbs]] of
													[]			= edge
													[x]			= x
													[_,x]		= x
													[_,_,x:_]	= x
where
	(compare,edge)								= if (home == West) (<,zero) (>,field.flength)

getOffsidePlayers :: !FootballField !Home !XPos ![Footballer] -> [FootballerID]
getOffsidePlayers field home line fbs			= map player_identity (filter (\fb=:{pos={px}} -> at_other_half px && (compare fb.pos.px line)) fbs)
where
	(compare,at_other_half)						= if (home == West) (>,(<) fieldhalf) (<,(>) fieldhalf)
	fieldhalf									= field.flength / 2.0
	
//	A referee can not see everything correct. 
filterOutMeanActions :: ![Footballer] ![FootballerWithEffect] !RandomSeed -> (![FootballerWithEffect],!RandomSeed)
filterOutMeanActions allPlayers allActions seed
	= ([action \\ action <- allActions & p <- ps | noticeAction action p],seed1)
where
	(ps,seed1)				= iterateStn (length allActions) nextRandomP seed
	
	noticeAction :: !FootballerWithEffect !P -> Bool
	noticeAction (fa,playerID) p
		| foul				= avg [p,1.0 - pa + if (isMember skill (skillsAsList fb)) 0.1 0.0] <= 0.5
		| otherwise			= True
	where
		fb					= find1 (identify_player playerID) allPlayers
		(foul,pa,skill)		= if (perhaps isTackled       fa) (True,chanceOfPenaltySuccess, Tackling)
							 (if (perhaps isSchwalbed     fa) (True,chanceOfSchwalbeSuccess,Schwalbing)
							 (if (perhaps isPlayedTheater fa) (True,chanceOfTheaterSuccess, PlayingTheater)
							 (if (perhaps isCaughtBall    fa) (True,chanceOfCatchSuccess,   Catching)
							                                  (False,undef,undef)
							 )))

nextBallPos :: !Position3D !Speed3D !(!Position,!XRadius,!ZRadius) -> Bool
nextBallPos item direction (target,radius,height)
							= isGettingToPos (iterate (next_pos direction) item)
where
	target_at_ground		= {zero & pxy=target}
	target_in_air			= {pxy=target, pz=height}
	
	next_pos :: !Speed3D !Position3D -> Position3D
	next_pos speed currentPosition
							= move_point3D {dxy={dx=(cos speed.vxy.direction)*new_v,dy=(sin speed.vxy.direction)*new_v},dz=new_v3} currentPosition
	where
		resistance			= if (currentPosition.pz > zero) air_resistance surface_resistance
		new_v				= resistance * speed.vxy.velocity
		new_v3				= speed.vz - 0.1*accelleration_sec
	
	isGettingToPos :: ![Position3D] -> Bool
	isGettingToPos []		= False
	isGettingToPos [x]		= False
	isGettingToPos [x:y:xs]
		| dist x target_at_ground < radius || dist x target_in_air < radius
							= True
		| avg [dist x target_at_ground,dist x target_in_air] < avg [dist y target_at_ground,dist y target_in_air]	// object is moving away
							= False
		| otherwise			= isGettingToPos [y:xs]																	// object is coming closer


//	Freekicks are granted from the pos of the player that is the most closest to the ball but not a position more forward to the goal
posForFreeKick :: !Home !Position /*!Position*/ ![Footballer] -> Position
posForFreeKick home ballPos /*victimPos*/ []				= ballPos//victimPos
posForFreeKick home ballPos /*victimPos*/ team
	= case [fb \\ fb <- team | (dist fb.pos ballPos) <= 15.0 ] of
		[]												= ballPos
		close											= (winner closer2ball close).pos
where
	closer2ball :: !Footballer !Footballer -> Footballer
	closer2ball fb1 fb2
		| dist fb1.pos ballPos < dist fb2.pos ballPos	= fb1
		| otherwise										= fb2

inKeeperSpot :: !Footballer !ClubName !Home !FootballField -> Bool
inKeeperSpot fb clubName home field	= inPenaltyArea field (if (getClubName fb == clubName) home (other home)) fb.pos
